/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "../../../glew/include/GL/glew.h"
#include "papermap.h"
#include "iOpenGLHelper.h"
#include <math.h>
#include <memory.h>
#include <QMutexLocker>


Papermap::Papermap (IOpenGLHelper * helper, int dpi) {
    m_Paper         = NULL;
    m_Bumpmap       = NULL;
    m_Helper        = helper;
    if (m_Helper) {
        m_Helper->addRef ();
    }

    m_Dpi = dpi;
}

Papermap::~Papermap (void) {
    SAFE_RELEASE (m_Bumpmap);
    SAFE_RELEASE (m_Helper);
}

void Papermap::setPaper (sPaper * map) {
    SAFE_RELEASE (m_Bumpmap);

    if (map) {
        m_Dpi = map->dpi;

        if (map->height) {
            m_Paper = map;
          //m_Paper->addRef ();     // FIXME Ignoring the reference number here.

            m_Bumpmap = g_fac_SPAWN_ARRAY (
                    m_Helper,
                    m_Paper->height->width (), m_Paper->height->height (),
                    m_Helper->proposeTextureFormat (OPENGLHELPER_TEXTURE_RGBA),
                    GL_FLOAT,
                    IArray::FLAGS_NO_SHARE | IArray::FLAGS_MIPMAPPING
            );

            generateBump ();
        }
    }
}

/// Generate a bump map.
void Papermap::generateBump (void) {
    QMutexLocker locker (m_Helper->mutex ());

    if (m_Bumpmap && m_Paper) {
              float * normal = (float*)        m_Bumpmap->lock_pointer (GL_WRITE_ONLY_ARB);
        const float * height = (const  float*) m_Paper->height->lock_pointer_const ();

        const int h = m_Paper->height->height ();
        const int w = m_Paper->height->width ();

        if (normal && height) {
            for (int y = 0; y < h; y++) {
                const int ym1 = (y + h - 1) % h;
                const int yp1 = (y + 1) % h;
                for (int x = 0; x < w; x++) {
                    const int xm1 = (x + w - 1) % w;
                    const int xp1 = (x + 1) % w;

                    const float dx = height[xp1 + y * w] - height[xm1 + y * w];
                    const float dy = height[x + yp1 * w] - height[x + ym1 * w];
                    const float m  = 1.0f / sqrtf (dx * dx + dy * dy + 1.0f);

                    float * const p = normal + (x + y * w) * 4;

                    p[0] = dx   * m;
                    p[1] = dy   * m;
                    p[2] = 1.0f * m;
                    p[3] = 0.0f;
                }
            }
        }

        m_Bumpmap->unlock ();
        m_Paper->height->unlock ();
    }
}

IArray * Papermap::detail (float detailLevel) const {
    if (!m_Paper)
        return NULL;

    int   idx  = -1;
    float mind = 100.0f;
    for (int i = 0; i < m_Paper->filtered_count; i++) {
        if (m_Paper->filtered[i].filtered) {
            const float d = fabsf (detailLevel - m_Paper->filtered[i].filter_r);
            if (mind > d) {
                mind = d;
                idx  = i;
            }
        }
    }

    if (idx >= 0) {
        return m_Paper->filtered[idx].filtered;
    } else {
        if (m_Paper) {
            return m_Paper->height;
        } else
            return NULL;
    }
}

IArray * Papermap::bump (void) const {
    if (m_Bumpmap) {
        return m_Bumpmap;
    } else {
        return NULL;
    }
}
