/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "connector.h"
#include <QDir>
#include <QApplication>
#include <QFileInfo>
#include <QStringList>
#include <QPluginLoader>
#include <QImage>
#include <QBuffer>
#include <QXmlSimpleReader>

#define ALLOW_INTERFACE_DECLARATION

#include "iInstrument.h"
#include "iPipeline.h"
#include "iRenderingChain.h"

#include "iPigmentDescription.h"
#include "pigment_xml.h"
#include "paper_xml.h"


IArray * iarrayFromFile (QString path, IOpenGLHelper * helper, int flags) {
    QImage i (path);
    if (i.isNull ())
        return NULL;

    IArray * a = g_fac_SPAWN_ARRAY (
            helper, i.width (), i.height (),
            helper->proposeTextureFormat (OPENGLHELPER_TEXTURE_LUMINANCE),
            GL_FLOAT,
            flags
    );
    if (a) {
        float * p = (float*) a->lock_pointer ();
        if (p) {
            for (int y = 0; y < i.height (); y++) {
                for (int x = 0; x < i.width (); x++, p++) {
                    *p = qGray (i.pixel (x, y)) / 255.0f;
                }
            }
        }

        a->unlock ();
    }

    return a;
}

bool iarrayToFileRgba (IArray * i, QString path) {
    const BYTE * data = (const BYTE *) i->lock_pointer_const ();
    if (data) {
        const int w = i->width ();
        const int h = i->height ();

        QImage img (w, h, QImage::Format_ARGB32);

        // Convert to QImage.
        const BYTE * src_line = data;
        for (int y = 0; y < h; y++, src_line += 4 * w) {
            QRgb * dst_line = (QRgb*) img.scanLine (y);
            for (int x = 0; x < w; x++) {
                dst_line[x] = qRgb (
                    src_line[x * 4 + 0],
                    src_line[x * 4 + 1],
                    src_line[x * 4 + 2]
                );
            }
        }

        // Save the image.
        i->unlock ();
        return img.save (path, NULL, 100);
    }

    return FALSE;
}

bool iarrayToFileGray (IArray * i, QString path) {
    const float * data = (const float *) i->lock_pointer_const ();
    if (data) {
        const int w = i->width ();
        const int h = i->height ();

        QImage img (w, h, QImage::Format_ARGB32);

        // Convert to QImage.
        const float * src_line = data;
        for (int y = 0; y < h; y++, src_line += w) {
            QRgb * dst_line = (QRgb*) img.scanLine (y);
            for (int x = 0; x < w; x++) {
                const BYTE px = (BYTE)(255.0f * src_line[x]);
                dst_line[x] = qRgb (
                    px, px, px
                );
            }
        }

        // Save the image.
        i->unlock ();
        return img.save (path, NULL, 100);
    }

    return FALSE;
}

IConnector * g_fac_SPAWN_CONNECTOR (IOpenGLHelper * helper) {
    return new Connector_loopback (helper);
}

Connector_loopback::Connector_loopback (IOpenGLHelper * helper) {
    m_Id             = ISpawnedObject::allocateId ();
    m_Helper         = helper;
    m_Logger         = NULL;
    m_LastError      = NULL;
    m_LoggingEnabled = FALSE;
    m_Connected      = FALSE;

    ISpawnedObject::_register (this);

    if (m_Helper) {
        m_Helper->addRef ();
    }
}

Connector_loopback::~Connector_loopback (void) {
    ISpawnedObject::_unregister (this);

  //disconnect ();

    // Release all instruments.
    foreach (IPlugin * plugin, m_Instruments) {
        plugin->release ();
    }
    foreach (IPlugin * plugin, m_Instruments_rep) {
        plugin->release ();
    }

    // Release all pipelines.
    foreach (IPlugin * pipe, m_Pipelines) {
        pipe->release ();
    }

    // Release all paper textures.
    foreach (sPaper * paper, m_PaperTextures) {
        SAFE_RELEASE (paper->color);
        SAFE_RELEASE (paper->height);
        for (int i = 0; i < paper->filtered_count; i++) {
            SAFE_RELEASE (paper->filtered[i].filtered);
        }

        delete paper;
    }

    // Release all pigments.
    foreach (IPigmentMixture * pig, m_Pigments) {
        pig->release ();
    }

    logMessage (LOGIMPORTANCE_NEVERMIND, "Connector is vanished in the air.");
    SAFE_RELEASE (m_Helper);
    SAFE_RELEASE (m_Logger);
}

void Connector_loopback::attachLogger (ILogger * logger) {
    SAFE_RELEASE (m_Logger);

    if (logger) {
        m_Logger = logger;
        m_Logger->addRef ();

        if (m_Helper)
            m_Helper->attachLogger (logger);

        // Notify all pipelines.
        foreach (IPlugin * pipe, m_Pipelines) {
            pipe->deInit ();
            ((IPipeline*)pipe)->attachLogger (logger);
        }
    }
}

void Connector_loopback::enableLogging (bool enable) {
    m_LoggingEnabled = enable;

    if (m_Helper)
        m_Helper->enableLogging (enable);

    // Notify all pipelines.
    foreach (IPlugin * pipe, m_Pipelines) {
        pipe->deInit ();
        ((IPipeline*)pipe)->enableLogging (enable);
    }
}

bool Connector_loopback::connect (const char * host, int port) {
    m_LastError = NULL;
    m_Connected = TRUE;

    Q_UNUSED (host);
    Q_UNUSED (port);

    QDir dir = QDir (DATA_PATH);

    QStringList paths;
    paths << dir.absoluteFilePath ("brushes");
    paths << dir.absoluteFilePath ("pipelines");

    loadPlugins    (paths);
    loadPaperpacks (dir.absoluteFilePath ("images/paper"));
    loadPigments   (dir.absoluteFilePath ("pigments"), &m_Pigments);

    // We are connected! 8)
    emit connected ();

    return TRUE;
}

bool Connector_loopback::disconnect (void) {
    m_LastError = NULL;
    m_Connected = FALSE;

    emit disconnected ();
    return TRUE;
}

void Connector_loopback::detachfromContext (void) {
    logMessage (LOGIMPORTANCE_NEVERMIND, "Purging to detach from the context.");

    // Notify all instruments.
    foreach (IPlugin * plugin, m_Instruments) {
        plugin->deInit ();
    }
    foreach (IPlugin * plugin, m_Instruments_rep) {
        plugin->deInit ();
    }

    // Notify all pipelines.
    foreach (IPlugin * pipe, m_Pipelines) {
        pipe->deInit ();
    }

    // Notify all paper textures.
    foreach (sPaper * paper, m_PaperTextures) {
        paper->height->purgeToSW ();
        paper->color->purgeToSW ();
        for (int i = 0; i < paper->filtered_count; i++) {
            paper->filtered[i].filtered->purgeToSW ();
        }
    }
}

void Connector_loopback::allocateObject (IConnectorCallback * caller, int param,
                                         eENTITY_TYPE objtype, const char * classname) {
    m_LastError = NULL;
    ISpawnedObject * entity = NULL;
    switch (objtype) {
        case ENTITY_RENDERING_CHAIN:
            entity = g_fac_SPAWN_RENDERING_CHAIN (m_Helper);
          //entity->addRef ();
            break;
        case ENTITY_INSTRUMENT_PLUGIN:
            // Return the whole pack at once.. >:-)
            foreach (IPlugin * plugin, m_Instruments) {
                if (classname) {
                    // Check the classname.
                    if (0 != strcmp (plugin->myClassName (), classname))
                        continue;
                }

                // Replicate...
                IInstrument * replica = dynamic_cast<IInstrument*>(plugin)->replicate ();
                m_Instruments_rep.append (replica);
                replica->addRef ();

                // ...and send.
                caller->complitedRequest (
                    param,
                    0,
                    replica
                );

                // FIXME Don't save the discarded replicas.
            }
            logMessage (LOGIMPORTANCE_NEVERMIND, "Sent the instruments pack to caller.");
            return;
        case ENTITY_PIPELINE_PLUGIN:
            // Return the whole pack at once.. >:-)
            foreach (IPlugin * plugin, m_Pipelines) {
                plugin->addRef ();

                // FIXME Object instance sharing?
                caller->complitedRequest (param, 0, plugin);
            }
            logMessage (LOGIMPORTANCE_NEVERMIND, "Sent the pipelines pack to caller.");
            return;
        case ENTITY_PIGMENT:
            // Return the whole pack at once.. >:-)
            foreach (IPigmentMixture * pig, m_Pigments) {
                pig->addRef ();

                // FIXME Object instance sharing?
                caller->complitedRequest (param, 0, pig);
            }
            logMessage (LOGIMPORTANCE_NEVERMIND, "Sent the pigments pack to caller.");
            return;
        case ENTITY_PAPER_TEXTURE:
            // Return the whole pack at once.. >:-)
            foreach (sPaper * paper, m_PaperTextures) {
                //paper->addRef ();

                // FIXME Object instance sharing?
                caller->complitedRequest (param, 0, paper);
            }
            logMessage (LOGIMPORTANCE_NEVERMIND, "Sent the paper textures pack to caller.");
            return;
    };

    if (entity) {
        logMessage (LOGIMPORTANCE_NEVERMIND, "Allocated an entity of type %s.", entity->myClassName ());
        caller->complitedRequest (param, 0, entity);
    } else {
        logMessage (LOGIMPORTANCE_ERROR, "Failed to allocate an entity of type %d.", objtype);
    }
}

void Connector_loopback::parsePigmentsPack (IConnectorCallback * caller, int param, const BYTE * data, int length) {
    QByteArray a (reinterpret_cast<const char*>(data), length);
    QBuffer buffer (&a);

    // Release all pigments.
    foreach (IPigmentMixture * pig, m_Pigments) {
        pig->release ();
    }

    // Load a new list.
    m_Pigments.clear ();
    if (loadPigments (&buffer, &m_Pigments)) {

        // Return the whole pack at once.. >:-)
        foreach (IPigmentMixture * pig, m_Pigments) {
            pig->addRef ();
            caller->complitedRequest (param, 0, pig);
        }
        logMessage (LOGIMPORTANCE_NEVERMIND, "Sent the pigments pack to caller.");
    }
}

bool Connector_loopback::callObject (int remoteId, int request, int length, const BYTE * data) {
    m_LastError = NULL;

    Q_UNUSED (remoteId);
    Q_UNUSED (request);
    Q_UNUSED (length);
    Q_UNUSED (data);

    return FALSE;
}

int Connector_loopback::waitObject (int remoteId, BYTE * data, int interval) {
    m_LastError = NULL;

    Q_UNUSED (remoteId);
    Q_UNUSED (data);
    Q_UNUSED (interval);

    return 0;
}

void Connector_loopback::loadPlugins (QStringList paths) {
    QStringList filters;

// FIXME More clever condition might be needed here.
#ifdef _WIN32
     filters << "*.dll";
#else
     filters << "*.so.*"
             << "*.so";
#endif

    QList<QObject*> all_plugins;

    int nstatic  = 0;
    int ndynamic = 0;
    logMessage (LOGIMPORTANCE_NEVERMIND, "Enumerating all static plugins ...");

   // Enumerate the static plugins.
    foreach (QObject * plugin, QPluginLoader::staticInstances ()) {
        all_plugins.append (plugin);
        nstatic++;
    }

    logMessage (LOGIMPORTANCE_NEVERMIND, "Found %d static plugins.", nstatic);

    foreach (QString path, paths) {
        QDir dir (path);
        QStringList files = dir.entryList (filters);
        logMessage (LOGIMPORTANCE_NEVERMIND, "Enumerating all plugins in folder %s ...",
                    path.toUtf8 ().data ());

        // Enumerate the dynamic plugins.
        foreach (QString file, files) {
            QString filepath = dir.absoluteFilePath (file);
            QFileInfo finfo (filepath);

            if (!finfo.isFile () || finfo.isSymLink ())
                continue;

            QPluginLoader lib (filepath);
            if (lib.load ()) {
                QObject * root = lib.instance ();
                if (root) {
                    all_plugins.append (root);
                    ndynamic++;
                } else {
                    logMessage (LOGIMPORTANCE_WARNING, "Failed to load file %s. Reason: %s.",
                                file.toUtf8 ().data (), lib.errorString ().toUtf8 ().data ());
                }
            } else {
                logMessage (LOGIMPORTANCE_WARNING, "Failed to load file %s. Reason: %s.",
                            file.toUtf8 ().data (), lib.errorString ().toUtf8 ().data ());

            }
        }
    }

    logMessage (LOGIMPORTANCE_NEVERMIND, "Found %d dynamic plugins.", ndynamic);

    // Try to load them all.
    foreach (QObject * plugin, all_plugins) {
        IPlugin * casted = NULL;

        // Probe IInstrumentFactory interface.
        if ((casted = qobject_cast<IInstrumentFactory*>(plugin))) {
            IInstrumentFactory * factory = dynamic_cast<IInstrumentFactory*>(casted);
            if (m_LoggingEnabled) {
                factory->attachLogger (m_Logger);
            }

            // Spawn the siblings.
            for (int i = 0; i < factory->instrumentsInLibrary (); i++) {
                IInstrument * instrument = factory->create (i, m_Helper);
                if (instrument) {
                    m_Instruments.append (instrument);
                    logMessage (LOGIMPORTANCE_NEVERMIND, "instrument sibling of class %s from %s.",
                             instrument->myClassName (), casted->myClassName ());
                }
            }
        } else
        // Probe IPipeline interface.
        if ((casted = qobject_cast<IPipelineFactory*>(plugin))) {
            IPipelineFactory * factory = dynamic_cast<IPipelineFactory*>(casted);
            if (m_LoggingEnabled) {
                factory->attachLogger (m_Logger);
            }

            // Spawn the siblings.
            for (int i = 0; i < factory->pipelinesInLibrary (); i++) {
                IPipeline * pipe = factory->create (i, m_Helper);
                if (pipe) {
                 // Find an apropriate position in queue.
                 QList<IPlugin*>::iterator inspos;
                 for (inspos = m_Pipelines.begin (); inspos != m_Pipelines.end (); inspos++)
                     if (((IPipeline*)(*inspos))->priority () <
                                            pipe->priority ())
                         break;

                 m_Pipelines.insert (inspos, pipe);
                 logMessage (LOGIMPORTANCE_NEVERMIND, "pipeline sibling of class %s from %s.",
                             pipe->myClassName (), casted->myClassName ());
                }
            }
        }

        if (casted) {
            logMessage (LOGIMPORTANCE_NEVERMIND, "Finished loading the entity of class %s.",
                        casted->myClassName ());

            //casted->release ();
        } else {
            logMessage (LOGIMPORTANCE_WARNING, "Failed to identify the plugin type.");// of %s.",
                //file.toUtf8 ().data ());
        }
    }
}

bool Connector_loopback::loadPaperpacks (QString path) {
    bool result = FALSE;
    QStringList filters;
                filters << "*.xml";

    QDir dir (path);
    QStringList files = dir.entryList (filters);

    // Enumerate all xml-files.
    foreach (QString file, files) {
        QString filepath = dir.absoluteFilePath (file);
        QFileInfo finfo (filepath);

        if (!finfo.isFile () || finfo.isSymLink ())
            continue;

        logMessage (LOGIMPORTANCE_NEVERMIND, "Enumerating all paper descriptions in file %s ...",
                filepath.toUtf8 ().data ());

        // Parse the file.
        QFile newfile (filepath);
        if (loadPaperpacks (&newfile)) {
            result = TRUE;
        } else {
            logMessage (LOGIMPORTANCE_WARNING, "Failed to parse %s ...",
                filepath.toUtf8 ().data ());
        }
    }

    return result;
}

bool Connector_loopback::loadPaperpacks (QIODevice * input) {
    // Release all paper textures.
    foreach (sPaper * paper, m_PaperTextures) {
        SAFE_RELEASE (paper->color);
        SAFE_RELEASE (paper->height);
        delete paper;
    }

    m_PaperTextures.clear ();

    bool result = FALSE;
    CQXmlHandler_Paper handler (m_Helper);

    QXmlSimpleReader reader;
    reader.setContentHandler (&handler);
    //reader.setErrorHandler   (&handler);

    QXmlInputSource src (input);
    if (reader.parse (src)) {
        QList<sPaper*> * pack = handler.papers ();
        if (pack) {
            foreach (sPaper * paper, *pack) {
                sPaper * paper1 = new sPaper;

                // Duplcate.
                paper1->height = paper->height;
                paper1->color  = paper->color;
                paper1->name   = paper->name;
                paper1->dpi    = paper->dpi;
                paper1->paper_class = paper->paper_class;

                paper1->filtered_count = paper->filtered_count;
                for (int i = 0; i < PREFILTERED_PAPER_COUNT; i++) {
                    paper1->filtered[i].filtered = NULL;
                    paper1->filtered[i].filter_r = 0.0f;
                }
                for (int i = 0; i < paper->filtered_count; i++) {
                    paper1->filtered[i].filtered = paper->filtered[i].filtered;
                    paper1->filtered[i].filter_r = paper->filtered[i].filter_r;
                    paper1->filtered[i].filtered->addRef ();
                }

                // Addref the duplicated data.
                paper1->height->addRef ();
                paper1->color->addRef ();

                // If there is no prefiltered data,
                // then generate it.
                if (paper1->filtered_count == 0) {
#ifndef GENERATE_MISSING_TEXTURES
                    logMessage (LOGIMPORTANCE_ERROR, "No prefiltered textures found. Most probably your installation is broken.");
#else
                    prefilteredGenerate (paper1);
#endif
                }

                m_PaperTextures.append (paper1);
                result = TRUE;
            }
        }
    }

    if (m_PaperTextures.size ()) {
        logMessage (LOGIMPORTANCE_NEVERMIND, "Successfully loaded %d textures.",
                    m_PaperTextures.size ());
    }

    // Reset the handler.
    handler.clear ();
    return result;
}

bool Connector_loopback::loadPigments (QString path, QList<IPigmentMixture*> * dest_list) {
    bool result = FALSE;
    QStringList filters;
                filters << "*.xml";

    QDir dir (path);
    QStringList files = dir.entryList (filters);

    // Enumerate all xml-files.
    foreach (QString file, files) {
        QString filepath = dir.absoluteFilePath (file);
        QFileInfo finfo (filepath);

        if (!finfo.isFile () || finfo.isSymLink ())
            continue;

        logMessage (LOGIMPORTANCE_NEVERMIND, "Enumerating all pigment descriptions in file %s ...",
                filepath.toUtf8 ().data ());

        // Parse the file.
        QFile newfile (filepath);
        if (loadPigments (&newfile, dest_list)) {
            result = TRUE;
        } else {
            logMessage (LOGIMPORTANCE_WARNING, "Failed to parse %s ...",
                filepath.toUtf8 ().data ());
        }
    }

    return result;
}

bool Connector_loopback::loadPigments (QIODevice * input, QList<IPigmentMixture*> * dest_list) {
    bool result = FALSE;
    CQXmlHandler_Pigment handler;

    QXmlSimpleReader reader;
    reader.setContentHandler (&handler);
    //reader.setErrorHandler   (&handler);

    QXmlInputSource src (input);
    if (reader.parse (src)) {
        QPIGMENTSHASH * pigpack = handler.pigments ();
        if (pigpack) {
            QHash<QString, QList<IPigmentDescription*>*>::const_iterator i = pigpack->constBegin ();
            while (i != pigpack->constEnd ()) {
                QString pack_name = i.key ();
                QList<IPigmentDescription*> * pack = i.value ();

                logMessage (LOGIMPORTANCE_NEVERMIND, "Loading pigment pack \'%s\' ...", pack_name.toUtf8 ().data ());
                int loaded = 0;
                for (int j = 0; j < pack->size (); j++) {
                    IPigmentDescription * pig = pack->at (j);

                    // Add the pigment to the list.
                    pig->addRef ();
                    dest_list->append (pig);
                    loaded++;
                }

                logMessage (LOGIMPORTANCE_NEVERMIND, "Parsed %d pigment descriptions.", loaded);
                i++;
            }

            result = TRUE;
        }
    }

    // Reset the handler.
    handler.clear ();
    return result;
}

void Connector_loopback::reloadShaders (void) {
    logMessage (LOGIMPORTANCE_NEVERMIND, "Reloading all shaders...");

    m_Helper->unloadShaders ();
    ISpawnedObject * o = ISpawnedObject::orphans ();

    // Enumerate all allocated objects.
    while (o) {
        if (0 == strcmp (o->myClassName(), "Server::Channel_gl")) {
            IChannel * c = dynamic_cast<IChannel*> (o);
            c->reloadShaders ();
        } else
        if (0 == strcmp (o->myClassName(), "Server::Pipeline_dry") ||
            0 == strcmp (o->myClassName(), "Server::Pipeline_advect")) {
            IPipeline * c = dynamic_cast<IPipeline*> (o);
            c->reloadShaders ();
        } else
        if (0 == strcmp (o->myClassName(), "Server::CRenderingChain_gl")) {
            IRenderingChain * c = dynamic_cast<IRenderingChain*> (o);
            c->reloadShaders ();
        }

        o = o->next ();
    }

    logMessage (LOGIMPORTANCE_NEVERMIND, "Done reloading all shaders.");
}

void Connector_loopback::prefilteredGenerate (sPaper * paper_d) {
    QMutexLocker locker (m_Helper->mutex ());

    IArrayFilter * filter = QHASMath::newGaussianBlur ();
    if (!filter)
        return;

    IArray * temp = g_fac_SPAWN_ARRAY (
        m_Helper,
        paper_d->height->width (), paper_d->height->height (),
        m_Helper->proposeTextureFormat (OPENGLHELPER_TEXTURE_LUMINANCE),
        GL_FLOAT,
        IArray::FLAGS_NO_SHARE
    );

    float r = 2;
    for (int i = 0; i < PREFILTERED_PAPER_COUNT; i++) {

        IArray * paper = paper_d->height;
        IArray * filtered =
        g_fac_SPAWN_ARRAY (
            m_Helper,
            paper->width (), paper->height (),
            m_Helper->proposeTextureFormat (OPENGLHELPER_TEXTURE_LUMINANCE),
            GL_FLOAT,
            IArray::FLAGS_NO_SHARE
        );

        SAFE_RELEASE (paper_d->filtered[i].filtered);
        paper_d->filtered[i].filtered = filtered;
        paper_d->filtered[i].filter_r = r;

        const int sz = paper->width () * paper->height ();

        // Duplicate.
        const float * src = (const float*) paper->lock_pointer_const ();
              float * dst =       (float*) filtered->lock_pointer ();

        memcpy (dst, src, sizeof(float) * sz);
        filtered->unlock ();
        paper->unlock ();

        // Filter the paper.
        filter->filter (filtered, temp, r);

        src = (const float*) paper->lock_pointer_const ();
        dst =       (float*) filtered->lock_pointer ();

        // Subtract and invert.
        float mi = 0.0f;//1.0f
        float ma = 0.0f;
        for (int i = 0; i < sz; i++) {
            dst[i] = 1.0f - MAX (0.0f, dst[i] - src[i]);
          //mi = MIN (mi, dst[i]);
            ma = MAX (ma, dst[i]);
        }

        // Re-normalize.
        float m = 1.0f / (ma - mi);
        for (int i = 0; i < sz; i++) {
            dst[i] = m * (dst[i] - mi);
        }

        filtered->unlock ();
        paper->unlock ();

        // Save the prefiltered texture,
        char suffix[128];
        snprintf (suffix, 128, " filtered-%.4d.bmp", (int)r);
        iarrayToFileGray (filtered, "filtered/" + paper_d->name + suffix);
        r *= 2;
    }

    paper_d->filtered_count = PREFILTERED_PAPER_COUNT;
    SAFE_RELEASE (temp);
    SAFE_RELEASE (filter);
}
