/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CONNECTOR_H
#define CONNECTOR_H

#include "commdefs.h"
#include "commtypes.h"
#include "refcounter.h"

#include "../../../glew/include/GL/glew.h"

#include "iConnector.h"
#include "iConnectorRequests.h"
#include "iOpenGLHelper.h"
#include "iPlugin.h"
#include "iPigmentMixture.h"
#include "iPaperSwitch.h"

#include <QObject>
#include <QDir>


class Connector_loopback : public QObject,
                           public IConnector,
                           public CRefCounter {
    Q_OBJECT
    _VERBOSEOBJECT

    int             m_Id;
    bool            m_Connected;
    IOpenGLHelper * m_Helper;

    /**
     * Loaded instrument plugins.
     */
    QList<IPlugin*> m_Instruments;

    /**
     * Saved pointers to the instrument replicas.
     */
    QList<IPlugin*> m_Instruments_rep;

    /**
     * Loaded pipeline plugins.
     */
    QList<IPlugin*> m_Pipelines;

    /**
     * Loaded pigment descriptors.
     */
    QList<IPigmentMixture*> m_Pigments;

    /**
     * Loaded paper textures.
     */
    QList<sPaper*>  m_PaperTextures;

    /**
     * Cached shared images.
     */
  //QList<IArray*>  m_SharedImages;

friend IConnector * g_fac_SPAWN_CONNECTOR (IOpenGLHelper * helper);
protected:

    Connector_loopback (IOpenGLHelper * helper);
    virtual ~Connector_loopback (void);

    /**
     * Load all available plugins.
     */
    void loadPlugins (QStringList path);

    /**
     * Load all available paper textures.
     */
    bool loadPaperpacks (QString path);
    bool loadPaperpacks (QIODevice * input);

    /**
     * Load all available pigment
     * descriptions.
     */
    bool loadPigments (QString path,      QList<IPigmentMixture*> * m_Pigments);
    bool loadPigments (QIODevice * input, QList<IPigmentMixture*> * m_Pigments);

    /**
     * Generate the sequence of prefiltered textures.
     */
    void prefilteredGenerate (sPaper * paper);

    virtual void suicide (void) {
        delete this;
    }
public:

    /******** ICONNECTOR IMPLANTS ********/

    /**
     * Connect to a server.
     */
    virtual bool connect (const char * host, int port);

    /**
     * Disconnect from a server.
     */
    virtual bool disconnect (void);

    /**
     * Get the current connection status.
     */
    virtual bool connected (void) const {
        return m_Connected;
    }

    /**
     * Request an entity allocation.
     */
    virtual void allocateObject (IConnectorCallback * caller, int param,
                                 eENTITY_TYPE objtype, const char * classname = NULL);

    /**
     * Parse the pigments descriprions and return
     * the pack via complitedRequest ().
     */
    virtual void parsePigmentsPack (IConnectorCallback * caller, int param, const BYTE * data, int length);

    /**
     * If connector is set into the blobcking
     * state it won't process any messages from
     * server, to ensure that waitObject () gets
     * them.
     */
    virtual void setBlobcking (bool blocking) {
        Q_UNUSED (blocking);
    }

    /**
     * Send a request to a remote object.
     * @param remoteId Id of the object on server side.
     * @param request Request code.
     * @param length Length of the input data in bytes.
     * @param data Input data buffer.
     * @return TRUE if the call was successful.
     */
    virtual bool callObject (int remoteId, int request, int length, const BYTE * data);

    /**
     * Wait for the response of a remote object.
     * @param remoteId Id of the object on server side.
     * @param data Output data buffer.
     * @param interval Number of milliseconds to wait.
     * @return The number of bytes recieved or 0 if error occured.
     */
    virtual int waitObject (int remoteId, BYTE * data, int interval = 1024);

    /**
     * Notify all objects that we're going
     * to lose the context.
     */
    virtual void detachfromContext (void);

    /**
     * Request all objects the reload their
     * shaders now.
     */
    virtual void reloadShaders (void);

    /******* ISPAWNEDOBJECT IMPLANTS *******/

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return "Client::Connector_loopback";
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

    /******* IVERBOSEOBJECT IMPLANTS *********/

    /**
     * Returns the description of last error
     * or NULL if no error occured.
     */
    virtual const char * lastError (void) const {
        return m_LastError;
    }

    /**
     * Attach the logger to an object.
     */
    virtual void attachLogger (ILogger * logger);

    /**
     * Enable of disable logging.
     */
    virtual void enableLogging (bool enable);

    /**
     * Check if the logging is enabled.
     */
    virtual bool loggingEnabled (void) {
        return m_LoggingEnabled;
    }

signals:
    void connected    (void);
    void disconnected (void);
};


#endif // CONNECTOR_H
