/***
 *  Copyright (C) 2009 Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "brushlist.h"

#define RQ_INSTRUMENTS_ENUM          0x0001



IInstrumentList * g_fac_SPAWN_INSTRUMENT_LIST (void) {
    return (IInstrumentList*) new Brushlist (NULL);
}

IInstrumentList * g_fac_SPAWN_INSTRUMENT_LIST (QWidget * parent) {
    return (IInstrumentList*) new Brushlist (parent);
}


Brushlist::Brushlist (QWidget *parent) : QFrame(parent),
    m_ui (new Ui::Brushlist) {
    m_ui->setupUi (this);

    m_Id = ISpawnedObject::allocateId ();

    m_Logger         = NULL;
    m_Connector      = NULL;
    m_Current        = NULL;
    m_LoggingEnabled = FALSE;
    m_LastPressure   = 0.0f;

    m_ui->comboBox->setIconSize(QSize(64, 16));
    m_ui->comboBox->view()->setIconSize(QSize(64, 64));
    static_cast<QListView*>(m_ui->comboBox->view())->setViewMode (QListView::IconMode);
    static_cast<QListView*>(m_ui->comboBox->view())->setMovement (QListView::Static);
    static_cast<QListView*>(m_ui->comboBox->view())->setFlow     (QListView::TopToBottom);

    QObject::connect (m_ui->comboBox,   SIGNAL(currentIndexChanged(int)), this, SLOT(instrumentListSelectionChanged(int)));
    QObject::connect (&m_ConfigPanel,   SIGNAL(instrumentSizeChanged(IInstrumentProxy*,int)),
                      this,             SIGNAL(instrumentSizeChanged(IInstrumentProxy*,int)));
    QObject::connect (m_ui->toolConfig, SIGNAL(toggled(bool)), this, SLOT(showConfigPanel(bool)));
    QObject::connect (&m_ConfigPanel,   SIGNAL(closed()), m_ui->toolConfig, SLOT(toggle()));

    m_StylusId = -1;
    m_ConfigPanel.hide ();
}

Brushlist::~Brushlist (void) {
    clear ();

    SAFE_RELEASE (m_Connector);
    SAFE_RELEASE (m_Logger);

    SAFE_DELETE (m_ui);
}

void Brushlist::connect (IConnector * caller) {
    SAFE_RELEASE (m_Connector);

    // Store the pointer and ask the
    // connector to deliver some plugins to us..
    if (caller) {
        m_Connector = caller;
        m_Connector->addRef ();
        m_Connector->allocateObject (this, RQ_INSTRUMENTS_ENUM, ENTITY_INSTRUMENT_PLUGIN);
    }
}

void Brushlist::complitedRequest (int param, int length, void * data) {
    Q_UNUSED (length);
    if (param == RQ_INSTRUMENTS_ENUM) {
        // We have one more instrument.. ^^'
        IInstrument * instrument = (IInstrument*) data;
        if (instrument) {
            // Accept only non-tool instruments.
            if (instrument->instrumentGroup () != INSTRUMENTGROUP_TOOLBOX) {
                append (instrument);
            }

            instrument->release ();
        }
    }
}

void Brushlist::clear (void) {
    // Release all instruments.
    foreach (IInstrumentProxy * instrument, m_Handles) {
        instrument->release ();
    }

    m_Handles.clear ();
    m_ui->comboBox->clear();
}

void Brushlist::changeEvent (QEvent *e) {
    switch (e->type()) {
    case QEvent::LanguageChange:
        m_ui->retranslateUi (this);
        break;
    default:
        break;
    }
}

void Brushlist::remove (IInstrument * target) {
    for (int i = 0; i < m_Handles.size (); i++) {
        IInstrumentProxy * instrument = m_Handles.at (i);
        if (instrument == target) {
            instrument->release ();
            m_Handles.removeAt (i);
            m_ui->comboBox->removeItem(i);
            return;
        }
    }
}

void Brushlist::removeAt (int index) {
    if (index >= 0 && index < m_Handles.size ()) {
        m_Handles.at (index)->release ();
        m_Handles.removeAt (index);
        m_ui->comboBox->removeItem(index);
    }
}

void Brushlist::append (IInstrument * instrument) {
    IInstrumentProxy * proxy = g_fac_SPAWN_INSTRUMENT_PROXY (instrument);
    if (proxy) {
        m_ui->comboBox->addItem(QIcon(QPixmap::fromImage(*reinterpret_cast<const QImage*>(proxy->qimage()))), "", qVariantFromValue (InstrumentPreview (proxy)));

        // Select the first row if there is
        // no current selection.
        if (m_Handles.size () == 0)
            m_ui->comboBox->setCurrentIndex(0);

        m_Handles.append (proxy);
    }
}

void Brushlist::prepend (IInstrument * instrument) {
    IInstrumentProxy * proxy = g_fac_SPAWN_INSTRUMENT_PROXY (instrument);
    if (proxy) {
        m_ui->comboBox->addItem(QIcon(QPixmap::fromImage(*reinterpret_cast<const QImage*>(proxy->qimage()))), "", qVariantFromValue (InstrumentPreview (proxy)));

        // Select the first row if there is
        // no current selection.
        if (m_Handles.size () == 0)
            m_ui->comboBox->setCurrentIndex(0);

        m_Handles.prepend (proxy);
    }
}

void Brushlist::promoteChannels (void) {
    // Broadcast all promoted channels.
    if (m_Current) {
        for (int i = 0; i < 64; i++) {
            FOURCC channel_uid = m_Current->enumPromotedChannels (i);
            if (channel_uid != FOURCC_NULL)
                emit channelPromoted (channel_uid);
            else
                break;
        }
    }
}

void Brushlist::canvasLayerChanged (ICanvasLayer * layer) {
    Q_UNUSED (layer);

    promoteChannels ();
}

void Brushlist::tabletPenChanged (QTabletEvent::PointerType pen) {
    m_StylusId = (int) pen;

    if (m_StatesHash.contains (pen)) {
        m_ui->comboBox->setCurrentIndex (m_StatesHash.value ((int)pen));
    } else if (m_Current) {
        m_StatesHash.insert ((int)pen, m_ui->comboBox->currentIndex ());
    }
}

void Brushlist::tabletDeviceChanged (qint64 uid) {
}

void Brushlist::instrumentListSelectionChanged (int row) {
    bool success = FALSE;
    m_Current = NULL;

    if (row < 0)
        return;

    QVariant data = m_ui->comboBox->itemData(row);
    if (qVariantCanConvert<InstrumentPreview> (data)) {
        InstrumentPreview pw = qVariantValue<InstrumentPreview> (data);

        // Detach the instrument
        if (m_Current)
            m_Current->detached ();

        m_Current = pw.instrument ();
        if (m_Current) {
            // Attach the instrument curve to a widget.
            m_ConfigPanel.setPressureCurve (m_Current->pressureCurve ());

            // Broadcast the new instrument.
            emit instrumentChanged     (m_Current);
            emit instrumentSizeChanged (m_Current,
                                        m_Current->currentPrimitiveWidth () / 2);

            // Attach it.
            m_Current->attached ();
            m_ConfigPanel.setInstrument (m_Current);

            // Broadcast all promoted channels.
            promoteChannels ();

            success = TRUE;
        } else {
            m_ConfigPanel.setPressureCurve (NULL);
        }
    }

    // Update the controls.
    m_ConfigPanel.rereadControls ();

    // Save the state.
    if (m_StylusId >= 0) {
        if (m_StatesHash.contains (m_StylusId)) {
            m_StatesHash.remove (m_StylusId);
        }

        m_StatesHash.insert (m_StylusId, row);
    }
}

void Brushlist::userStartsDrawing (void) {
}

void Brushlist::userStopsDrawing (void) {
}

void Brushlist::chageBrushScale (float d) {
    if (m_Current) {
        m_Current->setScaleModifier (m_Current->scaleModifier () + d);
        m_ConfigPanel.rereadControls ();
    }
}

void Brushlist::showConfigPanel (bool on) {
    if (on) {
        QPoint p = mapToGlobal (QPoint (0, 0));
        p.setY (p.y() + height () + 8);

        m_ConfigPanel.popup (p);
    } else {
        m_ConfigPanel.hide ();
    }
}

/// Repaint the item.
void InstrumentDelegate::paint (QPainter * painter, const QStyleOptionViewItem & option,
                                const QModelIndex & index) const {

    if (qVariantCanConvert<InstrumentPreview> (index.data ())) {
        // Retrieve the instrument data...
        InstrumentPreview pdata = qVariantValue<InstrumentPreview> (index.data ());
        painter->save ();

        // Mark the selection
        if (option.state & QStyle::State_Selected) {
            painter->fillRect (option.rect, QBrush (QColor (0, 255, 255, 255)));
        } else {
            painter->fillRect (option.rect, QBrush (QColor (200, 200, 200, 255)));
        }

        if (pdata.image ()) {
            // and paint it
            painter->drawImage (option.rect, * reinterpret_cast<const QImage*> (pdata.image ()));
        } else {
            painter->fillRect (option.rect, QBrush (QColor (0, 0, 0)));
        }

        painter->restore ();
    } else {
        QItemDelegate::paint (painter, option, index);
    }
}
