#ifndef CONTROLSSTACK_H
#define CONTROLSSTACK_H

#include "commdefs.h"
#include "commtypes.h"
#include "refcounter.h"

#include <QFrame>
#include <QPushButton>
#include <QList>
#include <QVBoxLayout>
#include <QScrollArea>

#include "iControlsStack.h"

class ControlsStack : public QFrame,
                      public IControlsStack,
                      public CRefCounter {
    Q_OBJECT;

    ControlsStack (QWidget * parent);
    virtual ~ControlsStack (void);

friend class QHASWidgets;
public:

    /********** ICONTROLSSTACK IMPLANTS *************/

    virtual QWidget * asWidget (void) {
        return this;
    }

    /**
     * Get the widget at
     * given position.
     */
    virtual QWidget * at (int index) {
        if (index >= 0 && index < m_Items.size ())
            return m_Items.at (index)->widget;
        else
            return NULL;
    }

    /**
     * Push the named widget into the stack.
     */
    virtual void push (QWidget * widget, QString name);

    /**
     * Switch the widget to be shown,
     */
    virtual void showWidget (QWidget * widget);

    /**
     * Switch the widget to be hidden,
     */
    virtual void hideWidget (QWidget * widget);

    /********* IVERBOSEOBJECT IMPLANTS *************/

    /**
     * Returns the description of last error
     * or NULL if no error occured.
     */
    virtual const char * lastError (void) const {
        return m_Error;
    }

    /**
     * Attach the logger to an object.
     */
    virtual void attachLogger (ILogger * logger) {
        SAFE_RELEASE (m_Logger);

        if (logger) {
            m_Logger = logger;
            m_Logger->addRef ();
        }
    }

    /**
     * Enable of disable logging.
     */
    virtual void enableLogging (bool enable) {
        m_LoggingEnabled = enable;
    }

    /**
     * Check if the logging is enabled.
     */
    virtual bool loggingEnabled (void) {
        return m_LoggingEnabled;
    }

    /********* ISPAWNEDOBJECT IMPLANTS **********/

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return "Client::ControlsStack";
    }

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

protected:
    struct sItem {
        QWidget     * widget;
        QPushButton * button;
    };

    ILogger     * m_Logger;
    QVBoxLayout * m_Box0;
    QVBoxLayout * m_Box1;
    QWidget     * m_Container;
    QScrollArea * m_Scroll;
    QWidget     * m_Spacer;
    QList<sItem*> m_Items;

    int           m_Id;
    bool          m_LoggingEnabled;
    const char *  m_Error;

    /**
     * A call to logger object.
     */
    void logMessage (eLOG_IMPORTANCE importance, const char * msg, ...) {
        if (msg && m_Logger && m_LoggingEnabled) {
            va_list val;
            va_start (val, msg);
            m_Logger->write (importance, this, msg, val);

            if (importance == LOGIMPORTANCE_ERROR ||
                importance == LOGIMPORTANCE_CRITICAL)
                m_Error = msg;
        }
    }

    virtual void suicide (void) {
        delete this;
    }

private slots:
    void triggered (bool checked);
};

#endif // CONTROLSSTACK_H
