#ifndef PAPERINDICATOR_H
#define PAPERINDICATOR_H

#include "commdefs.h"
#include "commtypes.h"
#include "refcounter.h"

#include "iPaperSwitch.h"
#include "iConnector.h"

#include <QtGui>


/**
 * This widget shows the current paper
 * texture and allows the user to change
 * it.
 */
class PaperIndicator : public QObject,
                       public IPaperSwitch,
                       public CRefCounter {
    Q_OBJECT
    _VERBOSEOBJECT

    PaperIndicator (QWidget * parent);
    virtual ~PaperIndicator (void);

friend class QHASWidgets;
public:

    virtual QWidget * asWidget (void) {
        return reinterpret_cast<QWidget*>(combo);
    }

    /**
     * Get the paper currently
     * selected in widget.
     */
    virtual sPaper * paper (void) {
        if (combo)
            return m_TexturesPack[combo->currentIndex ()];
        else
            return NULL;
    }

    /**
     * Find the paper with a given name
     * and set it as current.
     */
    virtual void setByName (QString name);

    /**
     * Rearrange the papers by their classes
     * and set the first one as current.
     */
    virtual void setByClass (sPaper::ePaperClass c);

    /******** IDATACONTAINER IMPLANTS ************/

    /**
     * Check if the container's content
     * was changed since the last commit.
     */
    virtual bool changesCheck (void) const {
        return FALSE;
    }

    /**
     * Commit the changes. This resets the
     * result returned by changesCheck ().
     */
    virtual void changesCommit (void) {}

    /**
     * The length of buffer required
     * to save the cantainer's data.
     */
    virtual unsigned long requiredBufferSize (void) const;

    /**
     * Save the container's data into
     * the external buffer.
     * @return Number of bytes actually written.
     */
    virtual unsigned long saveData (QDataStream * stream, unsigned long capacity);

    /**
     * Load the data from  the external
     * buffer into container.
     * @return Number of bytes actually read.
     */
    virtual unsigned long loadData (QDataStream * stream, unsigned long capacity);

    /******* ICONNECTORCALLBACK IMPLANTS *********/

    /**
     * Pass the connector interface
     * to the object.
     */
    virtual void connect (IConnector * caller);

    /**
     * A callback, called by connector when
     * the requested data is recieved.
     * @param param Caller-define parameter associated with request.
     * @param length The length of the data string in bytes.
     * @param data Data returned. This buffer will be invalid after the function exit.
     */
    virtual void complitedRequest (int param, int length, void * data);

    /********* IVERBOSEOBJECT IMPLANTS ***********/

    /**
     * Returns the description of last error
     * or NULL if no error occured.
     */
    virtual const char * lastError (void) const {
        return m_LastError;
    }

    /**
     * Attach the logger to an object.
     */
    virtual void attachLogger (ILogger * logger) {
        SAFE_RELEASE (m_Logger);

        if (logger) {
            m_Logger = logger;
            m_Logger->addRef ();
        }
    }

    /**
     * Enable or disable logging.
     */
    virtual void enableLogging (bool enable) {
        m_LoggingEnabled = enable;
    }

    /**
     * Check if the logging is enabled.
     */
    virtual bool loggingEnabled (void) {
        return m_LoggingEnabled;
    }

    /********* ISPAWNEDOBJECT IMPLANTS ***********/

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return "Client::PaperIndicator";
    }

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
        emit paperChanged (paper ());
    }

signals:
    /**
     * Emitted when user selects
     * the new paper texture.
     */
    void paperChanged (sPaper * paper);

private slots:
	// ComboBox was changed
    void comboChanged (int);

protected:
    //void paintEvent (QPaintEvent * e);

private:

    int          m_Id;
    IConnector * m_Connector;

    QComboBox * combo;

    /**
     * A list of all available
     * paper textures.
     */
    QList<sPaper*> m_TexturesPack;

    /**
     * Set a new paper texture.
     */
    void setPaper (sPaper * paper);

    /**
     * Deallocate all textures.
     */
    void clear (void);

    virtual void suicide (void) {
        delete this;
    }
};

#endif // PAPERINDICATOR_H
