/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "instrumentproxy.h"
#include "qhasmath.h"
#include <stdlib.h>



IInstrumentProxy * g_fac_SPAWN_INSTRUMENT_PROXY (IInstrument * instrument) {
    return (IInstrumentProxy*) new InstrumentProxy (instrument);
}


InstrumentProxy::InstrumentProxy (IInstrument * instrument) {
    m_Id = ISpawnedObject::allocateId ();
    m_PressureCurve = NULL;

    ISpawnedObject::_register (this);

    if (instrument) {
        m_Instrument = instrument;
        m_Instrument->addRef();

        double p0_pencil[7]  = {0.0f, 0.2f, 0.4f, 0.6f, 0.8f, 1.0f, 1.1f};
        double p1_pencil[7]  = {0.0f, 0.1f, 0.2f, 0.5f, 0.8f, 1.0f, 1.0f};
        double p0_default[7] = {0.0f, 0.2f, 0.4f, 0.6f, 0.8f, 1.0f, 1.1f};
        double p1_default[7] = {0.0f, 0.2f, 0.4f, 0.6f, 0.8f, 1.0f, 1.0f};

        /**
         * Individual pressure curves
         * for each instrument type.
         * FIXME Has to be adjusted later.
         */
        switch (m_Instrument->instrumentGroup ()) {
            case INSTRUMENTGROUP_PENCIL:
                m_PressureCurve = QHASMath::newSpline (p0_pencil, p1_pencil, 7);
            break;
            default:
                m_PressureCurve = QHASMath::newSpline (p0_default, p1_default, 7);
                break;
        }
    } else {
        m_Instrument = NULL;
    }

    m_Timeshot = 0;

    sState s;
    s.meter               = 0L;
    s.nQueued             = 0;
    s.nQueued_p           = 0;
    s.pressure            = 0.0f;
    s.velocity[0]         = 0.0f;
    s.velocity[1]         = 0.0f;

    // Push the initial state.
    m_States.push (s);

    m_Jitter              = 0.0f;
    m_ScaleModifier       = 1.0f;
    m_SensitivityModifier = 1.0f;

    meterReset ();
    anchoredMeterReset ();
}

InstrumentProxy::~InstrumentProxy (void) {
    ISpawnedObject::_unregister (this);

    SAFE_RELEASE (m_Instrument);
    SAFE_RELEASE (m_PressureCurve);
}

void InstrumentProxy::reset (void) {
    if (!m_States.isEmpty ()) {
        m_States.top ().nQueued     = 0;
        m_States.top ().nQueued_p   = 0;
        m_States.top ().meter       = 0L;
        m_States.top ().velocity[0] = 0.0f;
        m_States.top ().velocity[1] = 0.0f;
    }
}

void InstrumentProxy::setPosition (int x, int y, eInputMethod source) {
    if (m_States.isEmpty ())
        return;

    sState & s = m_States.top ();

    // Shift the queue.
    while (s.nQueued >= BRUSH_QUEUE_LENGTH) {
        for (int i = 0; i < s.nQueued - 1; i++) {
            s.xQueue[i] = s.xQueue[i + 1];
            s.yQueue[i] = s.yQueue[i + 1];
        }

        s.nQueued--;
    }

    // Update the meter.
    if (s.nQueued) {
        s.meter += abs (x - s.xQueue[s.nQueued - 1]) +
                   abs (y - s.yQueue[s.nQueued - 1]);
    }

    // Enqueue the position.
    s.xQueue[s.nQueued] = x;
    s.yQueue[s.nQueued] = y;
    s.nQueued++;

    clock_t t = clock ();
    if (m_Timeshot && s.nQueued > 1) {
        float dt = (t - m_Timeshot) / (float) CLOCKS_PER_SEC;
        if (IS_SIGNIFICANT (dt)) {
            const int dx = x - s.xQueue[s.nQueued - 2];
            const int dy = y - s.yQueue[s.nQueued - 2];
            float u  = dx / (100.0f * dt);
            float v  = dy / (100.0f * dt);

            // Clamp the velocity value.
            u = MATH::CLAMP (u, -10.0f, +10.0f);
            v = MATH::CLAMP (v, -10.0f, +10.0f);

            // Update the velocity values.
            s.velocity[0] = 0.9f * u + 0.1f * s.velocity[0];
            s.velocity[1] = 0.9f * v + 0.1f * s.velocity[1];
        }
    }

    m_Timeshot = t;

    if (m_Instrument)
        m_Instrument->setPosition (x, y, source);
}

void InstrumentProxy::setPressure (float pressure0, eInputMethod source) {
    if (m_States.isEmpty () || !m_Instrument)
        return;

    sState & s = m_States.top ();

    // Apply the pressure curve.
    float pressure;;
    if (m_PressureCurve)
        pressure = MATH::CLAMP_01 (m_PressureCurve->f (pressure0));
    else
        pressure = pressure0;

    // Accept.
    // Shift the queue.
    while (s.nQueued_p >= BRUSH_PRESS_QUEUE_LENGTH) {
        for (int i = 0; i < s.nQueued_p - 1; i++) {
            s.pressQueue[i] = s.pressQueue[i + 1];
        }

        s.nQueued_p--;
    }

    // Enqueue the pressure.
    s.pressQueue[s.nQueued_p] = pressure;
    s.nQueued_p++;

    s.pressure = pressure;
    if (m_Instrument)
        m_Instrument->setPressure (pressure, source);
}

void InstrumentProxy::stroke (float alpha, int & x, int & y) {
    if (m_States.isEmpty ())
        return;

    sState & s = m_States.top ();

#ifdef DEBUG_DISABLE_STROKE_INTERPOLATION
    if (TRUE) {
#else
    if (s.nQueued < 3) {
#endif
        x = s.xQueue[0] + (int) (m_Jitter * rand ()) / RAND_MAX;
        y = s.yQueue[0] + (int) (m_Jitter * rand ()) / RAND_MAX;
    } else {
        // Interpolate.
        BSpline::interpolate_quad (
            s.xQueue + s.nQueued - 3,
            s.yQueue + s.nQueued - 3,
            alpha
        );

        // Return the position with
        // the applied jitter.
        x = s.xQueue[s.nQueued] + (int) (m_Jitter * rand ()) / RAND_MAX;
        y = s.yQueue[s.nQueued] + (int) (m_Jitter * rand ()) / RAND_MAX;
    }
}

void InstrumentProxy::addMass (IPigmentMixture * pigment, float amount) {
    if (m_Instrument) {
        m_Instrument->addMass (pigment, amount);
    }
}

void InstrumentProxy::statePush (void) {
    if (!m_States.isEmpty ())
         m_States.push (m_States.top ());

    if (m_Instrument)
        m_Instrument->statePush ();
}

void InstrumentProxy::statePop (void) {
    if (m_States.size () > 1) {
        m_States.pop ();
    }

    if (m_Instrument)
        m_Instrument->statePop ();
}

void InstrumentProxy::attached (void) {
    if (m_Instrument) {
        if (COLORMETHOD_KMSAMPLE == m_Instrument->promoteColorChooser ()) {
            emit colorPromoted (
                    m_Instrument->currentAbsorbing (),
                    m_Instrument->currentScattering ()
            );
        }
    }
}

void InstrumentProxy::detached (void) {

}

void InstrumentProxy::anchoredMeterReset (void) {
    if (m_States.isEmpty ())
        return;

    sState & s = m_States.top ();

    if (s.nQueued) {
        s.anchor_x = s.xQueue[s.nQueued - 1];
        s.anchor_y = s.yQueue[s.nQueued - 1];
    } else {
        s.anchor_x = 0;
        s.anchor_y = 0;
    }
}

long InstrumentProxy::anchoredMeter (void) const {
    if (m_States.isEmpty ())
        return 0L;

    const sState & s = m_States.top ();

    if (s.nQueued) {
        // Manhattan distance.
        return abs (s.anchor_x - s.xQueue[s.nQueued - 1]) +
               abs (s.anchor_y - s.yQueue[s.nQueued - 1]);
    } else {
        return 0L;
    }
}
