/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "newpainting.h"
#include "preferences.h"
#include "preferencesdialog.h"

MainWindow::MainWindow (QWidget *parent, Logger_signal * logger) :
    QMainWindow(parent),
    ui (new Ui::MainWindowClass),
    m_APP_TITLE (tr ("QAquarelle"))
{
    ui->setupUi (this);
    ui->colorPicker->setPickerStyle (PREF.colorPickerStyle);

    QHASWidgets::init ();
    QDir mypath (DATA_PATH);

    m_WheelCurrentModifier = WheelDoesNothing;

    // Spawn and connect the logger.
    m_Logger = logger;
    if (m_Logger)
        m_Logger->addRef ();

    connect (logger, SIGNAL(message(eLOG_IMPORTANCE, const char*)),
             this,   SLOT(logmessage(eLOG_IMPORTANCE, const char *)));

    m_Helper = g_fac_SPAWN_OPENGL_HELPER ();
    m_Helper->attachLogger   (m_Logger);
    m_Helper->enableLogging  (TRUE);
    m_Helper->setShadersPath (mypath.absoluteFilePath ("shaders").toUtf8 ().data ());
    m_Helper->setPerformanceParameter (OPENGLHELPER_TEX_COMPRESSION,
                                       PREF.ogl_textureCompression);

    m_Worker = new Worker ();

    // Load the skin data.
    m_Skin = g_fac_SPAWN_SKIN (mypath.absoluteFilePath ("images/gui"));

    m_wrkCanvasIterate       = NULL;
    m_wrkCanvasPreview       = NULL;
    m_HasGLContext           = FALSE;
    colorChangedByMainWindow = FALSE;
    m_RestoreInstrumentColor = TRUE;
    m_UpdateBeep             = FALSE;
    m_Block                  = FALSE;

    m_MdiSub         = NULL;
    m_Workspace      = NULL;
    m_Connector      = g_fac_SPAWN_CONNECTOR (m_Helper);
    m_Canvas         = NULL;
  //m_Mixer          = NULL;
    m_InstrumentList = g_fac_SPAWN_INSTRUMENT_LIST     (this);
    m_LayersList     = NULL;
    m_PaperSwitch    = QHASWidgets::spawnPaperSwitch   (ui->dockCanvas);
    m_Palette        = g_fac_SPAWN_PALETTE             (ui->dockPalette, m_Skin, TRUE);
    m_dynPalette     = g_fac_SPAWN_PALETTE             (ui->dockPalette, m_Skin, FALSE);

    QTabWidget *palettetab = new QTabWidget(this);
    ui->dockPalette->setWidget (palettetab);
    palettetab->addTab(m_dynPalette->asWidget(), tr("Cached colors"));
    palettetab->addTab(m_Palette->asWidget (), tr("Palette"));
    QObject::connect(ui->colorPicker, SIGNAL(pickUpColorExt(IPigmentMixture *)), dynamic_cast<QObject*>(m_dynPalette), SLOT(insertCopyOfPigment(IPigmentMixture*)));

    // Init some docks.
    ui->dockCanvas->setWidget (m_PaperSwitch->asWidget ());

    m_Connector->attachLogger  (m_Logger);
    m_Connector->enableLogging (TRUE);
    m_InstrumentList->attachLogger (m_Logger);
    m_InstrumentList->enableLogging (TRUE);
    if (m_LayersList) {
        m_LayersList->attachLogger (m_Logger);
        m_LayersList->enableLogging (TRUE);

        connect (dynamic_cast<QObject*>(m_LayersList),
                 SIGNAL(layerChanged(ICanvasLayer*)),
                 dynamic_cast<QObject*>(m_InstrumentList),
                 SLOT(canvasLayerChanged(ICanvasLayer*))
        );
    }

    connect(ui->paintingStyle, SIGNAL(paintingStyleChanged()), this, SLOT(paintingStyleChanged()));

    connect (dynamic_cast<QObject*>(m_Connector), SIGNAL(connected()),    this, SLOT(connected()));
    connect (dynamic_cast<QObject*>(m_Connector), SIGNAL(disconnected()), this, SLOT(disconnected()));
    connect (dynamic_cast<QObject*>(m_PaperSwitch),
             SIGNAL(paperChanged (sPaper*)),
             m_Worker,
             SLOT(paperChanged (sPaper*))
    );
    connect (dynamic_cast<QObject*>(m_InstrumentList),
             SIGNAL(instrumentChanged (IInstrumentProxy*)),
             this,
             SLOT(instrumentChanged (IInstrumentProxy*))
    );
    connect (ui->toolPanel,
             SIGNAL(instrumentChanged(IInstrumentProxy*)),
             this,
             SLOT(toolInstrumentChanged(IInstrumentProxy*))
    );

    // Attach the color picker.
    connect (ui->colorPicker,
             SIGNAL(colorChangedExt(IPigmentMixture*, float, eColorMetod)),
             this,
             SLOT(colorChangedByColorPicker(IPigmentMixture*, float, eColorMetod))
    );

    // ...and palette too.
    connect (dynamic_cast<QObject*>(m_Palette),
             SIGNAL(touched (IPigmentMixture*, float, eColorMetod)),
             this,
             SLOT(colorChanged(IPigmentMixture*, float, eColorMetod))
    );
    connect (dynamic_cast<QObject*>(m_dynPalette),
             SIGNAL(touched (IPigmentMixture*, float, eColorMetod)),
             this,
             SLOT(colorChanged(IPigmentMixture*, float, eColorMetod))
    );

    disconnected ();

    if (m_Connector->connect ("127.0.0.1", CONNECTOR_PORT)) {
        m_InstrumentList->connect (m_Connector);
        m_PaperSwitch->connect    (m_Connector);
        m_Palette->connect        (m_Connector);
      //m_dynPalette->connect     (m_Connector);
        ui->toolPanel->connect    (m_Connector);
    }

    ui->statusBar->insertPermanentWidget (0, ui->frameStatus, 1);

    // Create the MDI area.
    m_MdiArea = new QMdiArea;
    m_MdiArea->setHorizontalScrollBarPolicy (Qt::ScrollBarAsNeeded);
    m_MdiArea->setVerticalScrollBarPolicy   (Qt::ScrollBarAsNeeded);
    //m_MdiArea->setBackground (  FIXME Skinned background.

    // Attach MDI area to the form.
    //m_layoutCanvas->addWidget (m_MdiArea);
    setCentralWidget(m_MdiArea);

    // Init the actions.
    connect (ui->actionStart_Simulation,        SIGNAL (triggered()), this, SLOT (startSimulation()));
    connect (ui->actionStop_Simulation,         SIGNAL (triggered()), this, SLOT (stopSimulation()));
    connect (ui->actionSave_as,                 SIGNAL (triggered()), this, SLOT (saveAs()));
    connect (ui->actionSave,                    SIGNAL (triggered()), this, SLOT (save()));
    connect (ui->actionOpen,                    SIGNAL (triggered()), this, SLOT (open()));
    connect (ui->actionNew,                     SIGNAL (triggered()), this, SLOT (create()));
    connect (ui->actionAbout,                   SIGNAL (triggered()), this, SLOT (about()));
    connect (ui->actionAbout_Qt4,               SIGNAL (triggered()), qApp, SLOT (aboutQt()));
    connect (ui->actionExit,                    SIGNAL (triggered()), this, SLOT (close()));
    connect (ui->actionClose,                   SIGNAL (triggered()), this, SLOT (clean()));
    connect (ui->actionClear_canvas,            SIGNAL (triggered()), this, SLOT (clearCanvas()));
    connect (ui->actionDry_canvas,              SIGNAL (triggered()), this, SLOT (dryCanvas()));
    connect (ui->actionExport_as,               SIGNAL (triggered()), this, SLOT (notAssigned()));
    connect (ui->actionRedo,                    SIGNAL (triggered()), this, SLOT (notAssigned()));
    connect (ui->actionUndo,                    SIGNAL (triggered()), this, SLOT (notAssigned()));
    connect (ui->actionReload_shaders,          SIGNAL (triggered()), this, SLOT (reloadShaders()));
    connect (ui->actionShow_debug_log,          SIGNAL (triggered()), this, SLOT (notAssigned()));

    connect (ui->spinScale, SIGNAL (valueChanged(double)), this, SLOT (changeCanvasScale(double)));

  //connect (ui->dockInstruments, SIGNAL (visibilityChanged(bool)), this, SLOT (dockVisibility(bool)));
    connect (ui->dockPicker,      SIGNAL (visibilityChanged(bool)), this, SLOT (dockVisibility(bool)));
    connect (ui->dockCanvas,      SIGNAL (visibilityChanged(bool)), this, SLOT (dockVisibility(bool)));

    ui->actionStart_Simulation->setEnabled (FALSE);
    ui->actionStop_Simulation->setEnabled  (TRUE);

    ui->actionSave->setEnabled     (FALSE);
    ui->actionOpen->setEnabled     (FALSE);
    ui->actionNew->setEnabled      (FALSE);
    ui->actionSave_as->setEnabled  (FALSE);
    ui->actionClose->setEnabled    (FALSE);

    m_Tracked        = FALSE;
    m_LastTrackX     = 0;
    m_LastTrackY     = 0;

    resetDragStates ();
    createShortcuts ();
    createHotkeys ();
    createToolbars ();
    createCanvas ();
    showMaximized ();

    setWindowTitle (m_APP_TITLE);

    QMenu * cm = createPopupMenu();
    cm->setTitle(tr("Panels"));
    ui->menuWindow->addMenu(cm);

    connect (&m_TimerConsciousness, SIGNAL (timeout()), this, SLOT (consciousness()));
    m_TimerConsciousness.start (512);

    m_Block = TRUE;
    ui->actionPaint_bump_mapping->setChecked (PREF.render_paintBump);
    ui->actionPaper_bump_mapping->setChecked (PREF.render_paperBump);
    ui->actionMotion_vectors->setChecked (PREF.render_mvectors);
    m_Block = FALSE;
}

MainWindow::~MainWindow (void) {
    // Terminate all works.
    if (m_Worker)
        m_Worker->terminate ();

    // Close the workspace if any.
    closeWorkspace ();

    SAFE_DELETE (m_Worker);

    //SAFE_RELEASE (m_Mixer);
    SAFE_RELEASE (m_wrkCanvasIterate);
    SAFE_RELEASE (m_InstrumentList);
    SAFE_RELEASE (m_LayersList);
    SAFE_RELEASE (m_PaperSwitch);
    SAFE_RELEASE (m_Palette);
    SAFE_RELEASE (m_dynPalette);
    SAFE_RELEASE (m_Skin);

    SAFE_DELETE (m_MdiArea);

    // Purge all OpenGL-related objects.
    m_Connector->detachfromContext ();
    SAFE_RELEASE (m_Connector);

    SAFE_RELEASE (m_Canvas);

    m_Helper->deinit ();
    SAFE_RELEASE (m_Helper);
    SAFE_RELEASE (m_Logger);
    delete ui;
}
/*
void MainWindow::setStatusTip (const QString & tip) {
    ui->txtMessage->setText (tip);
}

QString MainWindow::statusTip (void) const {
    return ui->txtMessage->text ();
}
*/
void MainWindow::createToolbars (void) {
    // Load the icons from the skin.
    ui->actionNew->setIcon               (QIcon (m_Skin->pixmapPathByName ("icons/icon_new.png")));
    ui->actionOpen->setIcon              (QIcon (m_Skin->pixmapPathByName ("icons/icon_open.png")));
    ui->actionSave->setIcon              (QIcon (m_Skin->pixmapPathByName ("icons/icon_save.png")));
    ui->actionStart_Simulation->setIcon  (QIcon (m_Skin->pixmapPathByName ("icons/icon_start.png")));
    ui->actionStop_Simulation->setIcon   (QIcon (m_Skin->pixmapPathByName ("icons/icon_stop.png")));
    ui->actionDry_canvas->setIcon        (QIcon (m_Skin->pixmapPathByName ("icons/icon_dry.png")));
    ui->actionClear_canvas->setIcon      (QIcon (m_Skin->pixmapPathByName ("icons/icon_clear.png")));

    m_FileToolBar = addToolBar (tr ("File"));
    m_FileToolBar->setObjectName ("File Toolbar");
    m_FileToolBar->addAction (ui->actionNew);
    m_FileToolBar->addAction (ui->actionOpen);
    m_FileToolBar->addAction (ui->actionSave);
    m_FileToolBar->setIconSize (QSize (16, 16));

    m_EditToolBar = addToolBar (tr ("Edit"));
    m_EditToolBar->setObjectName ("Edit Toolbar");
    m_EditToolBar->addAction (ui->actionStart_Simulation);
    m_EditToolBar->addAction (ui->actionStop_Simulation);
    m_EditToolBar->setIconSize (QSize (16, 16));

    m_WaterToolBar = addToolBar (tr ("Water Control"));
    m_WaterToolBar->setObjectName ("Water Toolbar");
    m_WaterToolBar->addWidget (ui->paintingStyle);
    m_WaterToolBar->addAction (ui->actionDry_canvas);
    m_WaterToolBar->addAction (ui->actionClear_canvas);
    m_WaterToolBar->addWidget (m_InstrumentList->asWidget ());
    m_WaterToolBar->setIconSize (QSize (16, 16));
}

void MainWindow::createCanvas (void) {
    // Create the canvas object.
    m_Canvas = g_fac_SPAWN_CANVAS (NULL, 800, 600, m_Helper);
    if (m_Canvas) {
        m_Canvas->asWidget ()->setVisible (FALSE);
        m_Canvas->attachLogger  (m_Logger);
        m_Canvas->enableLogging (TRUE);

        // And setup some connections...
        // Arrrggghh...
        connect (dynamic_cast<QObject*>(m_Canvas),    SIGNAL(userStartsPainting()),     this,       SLOT(userStartedPainting()));
        connect (dynamic_cast<QObject*>(m_Canvas),    SIGNAL(userStartsPainting()),     m_Worker,   SLOT(userStartedPainting()));
        connect (dynamic_cast<QObject*>(m_Canvas),    SIGNAL(userStopsPainting()),      this,       SLOT(userStoppedPainting()));
        connect (dynamic_cast<QObject*>(m_Canvas),    SIGNAL(userStopsPainting()),      m_Worker,   SLOT(userStoppedPainting()));
        connect (dynamic_cast<QObject*>(m_Canvas),    SIGNAL(glcontextObtained()),      this,       SLOT(glContextGained()));
        connect (dynamic_cast<QObject*>(m_Canvas),    SIGNAL(glcontextLost()),          this,       SLOT(glContextLost()));
        connect (dynamic_cast<QObject*>(m_Canvas),    SIGNAL(renderingOptsRequest()),   this,       SLOT(renderingOptsRequested()));
        connect (dynamic_cast<QObject*>(m_Canvas),
                 SIGNAL(tracking(float, float, float, float, float, bool)),
                 this,
                 SLOT(canvasTrack(float, float, float, float, float, bool))
        );
        connect (dynamic_cast<QObject*>(m_Connector),
                 SIGNAL(disconnected()),
                 dynamic_cast<QObject*>(m_Canvas),
                 SLOT(disconnected())
        );
        connect (dynamic_cast<QObject*>(m_InstrumentList),
                 SIGNAL(instrumentChanged (IInstrumentProxy*)),
                 dynamic_cast<QObject*>(m_Canvas),
                 SLOT(instrumentChanged (IInstrumentProxy*))
        );
        connect (dynamic_cast<QObject*>(m_InstrumentList),
                 SIGNAL(instrumentSizeChanged (IInstrumentProxy*, int)),
                 dynamic_cast<QObject*>(m_Canvas),
                 SLOT(instrumentSizeChanged (IInstrumentProxy*, int))
        );
        connect (dynamic_cast<QObject*>(m_InstrumentList),
                 SIGNAL(channelPromoted(FOURCC)),
                 dynamic_cast<QObject*>(m_Canvas),
                 SLOT(channelPromoted(FOURCC))
        );
        connect (dynamic_cast<QObject*>(m_Canvas),
                 SIGNAL(tabletPenChanged(QTabletEvent::PointerType)),
                 dynamic_cast<QObject*>(m_InstrumentList),
                 SLOT(tabletPenChanged(QTabletEvent::PointerType))
        );
        connect (dynamic_cast<QObject*>(m_Canvas),
                 SIGNAL(tabletDeviceChanged (qint64)),
                 dynamic_cast<QObject*>(m_InstrumentList),
                 SLOT(tabletDeviceChanged (qint64))
        );
        connect (dynamic_cast<QObject*>(m_Canvas),
                 SIGNAL(strokeTurns()),
                 this,
                 SLOT(strokeDirectionChanged())
        );
        connect (dynamic_cast<QObject*>(m_PaperSwitch),
                 SIGNAL(paperChanged (sPaper*)),
                 dynamic_cast<QObject*>(m_Canvas),
                 SLOT(paperChanged (sPaper*))
        );
        connect (m_Worker,
                 SIGNAL(statusChanged (QHASGlobal::eStatus)),
                 dynamic_cast<QObject*>(m_Canvas),
                 SLOT(statusChanged (QHASGlobal::eStatus))
        );
        connect (m_Worker,
                 SIGNAL(statusChanged (QHASGlobal::eStatus)),
                 this,
                 SLOT(statusChanged (QHASGlobal::eStatus))
        );
        connect (dynamic_cast<QObject*>(m_Canvas),
                 SIGNAL(colorPicked (IPigmentMixture*)),
                 this,
                 SLOT(colorChanged(IPigmentMixture*))
        );
        //
        // Woof...


        if (m_Connector) {
            if (m_Connector->connected ()) {
                m_Canvas->connect (m_Connector);
            }
        }

        // FIXME Here I try to force the initializeGL call.
        dynamic_cast<QGLWidget*>(m_Canvas->asWidget ())->updateGL ();

        if (PREF.pressureControl)
            m_Canvas->setPressureWorkingRange(PREF.pressureMin, PREF.pressureMax);
        else
            m_Canvas->setPressureWorkingRange(0.f, 1.f);

        // And gracefully fail if there was an error.
        if (!m_HasGLContext) {
            QMessageBox::warning (
                    this,
                    tr ("QAquarelle"),
                    tr ("I'm sorry, but either I couldn't initialize the OpenGL on your system "
                        "or there's no support for some important extensions... "
                        "I cannot proceed any further for it is complitely useless now.\n\n"
                        "Try to upgrade your videocard driver if it is too old.\n"
                        "Underlying system returns the following error: %1")
                    .arg (m_Helper->lastError ())
            );

            statusBar()->showMessage (tr ("No OpenGL context."), STATUSBAR_MESSAGE_TIMEOUT);
            statusChanged (QHASGlobal::STATUS_DEAD);
        } else {
            statusBar()->showMessage (tr ("Ready."), STATUSBAR_MESSAGE_TIMEOUT);
            statusChanged (QHASGlobal::STATUS_IDLE);
        }
    }
}

void MainWindow::createShortcuts (void) {
    m_Shortcut_ClearCanvas = new QShortcut (this);

    connect (m_Shortcut_ClearCanvas, SIGNAL(activated()), this, SLOT (clearCanvas()));
    m_Shortcut_ClearCanvas->setWhatsThis (tr ("Clear canvas"));
}

void MainWindow::createHotkeys (void) {
    m_ModifierKeys.clear ();

    // NOTE Use QShortcut::whatsThis() to display it in a dialog.

    // Init the modifiers.
	m_ModifierKeys.clear();
    for(int i=0; i<DRAGSTATE_COUNT; i++) {
        if(PREF.dragKeys[i] >= 0)
            m_ModifierKeys.insert (PREF.dragKeys[i], (eDragMode)i);
    }
	m_ModifierKeysWheel.clear();
    for(int i=0; i<WheelCOUNT; i++) {
        if(PREF.wheelKeys[i] >= 0)
            m_ModifierKeysWheel.insert (PREF.wheelKeys[i], (WheelAction)i);
    }

    // Default actions.
	if(m_ModifierKeysWheel.contains (0)) {
		m_WheelCurrentModifier = m_ModifierKeysWheel.value(0);
	} else {
		m_WheelCurrentModifier = WheelDoesNothing;
	}

    // Init the shortcuts.
    m_Shortcut_ClearCanvas->setKey (QKeySequence ("Alt+Shift+C"));
}

void MainWindow::glContextGained (void) {
    m_HasGLContext = TRUE;
    if (m_InstrumentList)
        m_InstrumentList->promoteChannels ();

    if (m_Canvas && m_LayersList) {
        // FIXME Unneeded?
        m_LayersList->attach (m_Canvas);
        m_Canvas->asWidget ()->update ();
    }

    ui->actionOpen->setEnabled     (TRUE);
    ui->actionNew->setEnabled      (TRUE);

    statusBar()->showMessage (tr ("Gained OpenGL context."), STATUSBAR_MESSAGE_TIMEOUT);
}

void MainWindow::glContextLost (void) {
    m_HasGLContext = FALSE;

    statusBar()->showMessage (tr ("Lost OpenGL context."), STATUSBAR_MESSAGE_TIMEOUT);
}

void MainWindow::canvasTrack (float x,    float y,
                              float dxdt, float dydt,
                              float pressure,
                              bool  inside) {

    QFont font = ui->txtXPos->font ();
    font.setBold (inside);

    char buf[128];
    ui->txtXPos->clear ();
    ui->txtXPos->setFont (font);
    ui->txtYPos->clear ();
    ui->txtYPos->setFont (font);
    ui->txtPressure->clear ();
    ui->txtPressure->setFont (font);

    snprintf (buf, 128, "%.1f%+.1f", x, dxdt);
    ui->txtXPos->setText (buf);

    snprintf (buf, 128, "%.1f%+.1f", y, dydt);
    ui->txtYPos->setText (buf);

    if (IS_SIGNIFICANT (pressure)) {
        snprintf (buf, 128, "%+.2f", pressure);
    } else {
        buf[0] = 0;
    }
    ui->txtPressure->setText (buf);

    m_LastTrackX = (int) x;
    m_LastTrackY = (int) y;
    m_Tracked    = TRUE;
}

void MainWindow::statusChanged (QHASGlobal::eStatus status) {
    switch (status) {
        case QHASGlobal::STATUS_DEAD:
            ui->iconStatus->iconChanged (m_Skin->pixmapByName ("icons/icon_st_dead.png"));
            break;
        case QHASGlobal::STATUS_IDLE:
            ui->iconStatus->iconChanged (m_Skin->pixmapByName ("icons/icon_st_idle.png"));
            break;
        case QHASGlobal::STATUS_PRIORITY_SYSTEM:
            ui->iconStatus->iconChanged (m_Skin->pixmapByName ("icons/icon_st_system.png"));
            break;
        case QHASGlobal::STATUS_PRIORITY_USER:
            ui->iconStatus->iconChanged (m_Skin->pixmapByName ("icons/icon_st_user.png"));
            break;
    }
}

void MainWindow::logmessage (eLOG_IMPORTANCE importance, const char * str) {
    // Duplicate the errors into stderr.
    if (importance != LOGIMPORTANCE_NEVERMIND)
        fprintf (stderr, "%s", str);
}

void MainWindow::connected (void) {
    ui->txtConnection->setText (tr ("connected"));
}

void MainWindow::disconnected (void) {
    ui->txtConnection->setText (tr ("disconnected"));
}

bool MainWindow::isSimulationStarted (void) const {
    if (m_wrkCanvasIterate) {
        return m_wrkCanvasIterate->isActive ();
    } else
        return FALSE;
}

void MainWindow::startSimulation (void) {
    ui->actionStart_Simulation->setEnabled (FALSE);
    ui->actionStop_Simulation->setEnabled  (TRUE);

    if (m_wrkCanvasIterate)
        m_wrkCanvasIterate->setActive (TRUE);
}

void MainWindow::stopSimulation (void) {
    ui->actionStart_Simulation->setEnabled (TRUE);
    ui->actionStop_Simulation->setEnabled  (FALSE);

    if (m_wrkCanvasIterate)
        m_wrkCanvasIterate->setActive (FALSE);
}

/// Dispatch the color changing event
/// to the instruments.
void MainWindow::colorChangedByColorPicker (IPigmentMixture * pig, float amount, eColorMetod source) {
    Q_UNUSED (source);
    if (pig && m_InstrumentList && m_HasGLContext) {
        IInstrumentProxy * i = m_InstrumentList->current ();
        if (i) {
            // Apply the color.
            i->addMass (pig, amount);
        }
    }
}

void MainWindow::colorChanged (IPigmentMixture * pig) {
    colorChanged (pig, 1.0f, COLORMETHOD_KMSAMPLE);
}

/// Dispatch the color changing event
/// to the instruments.
void MainWindow::colorChanged (IPigmentMixture * pig, float amount, eColorMetod source) {
    Q_UNUSED (source);
    if (!m_InstrumentList)
        return;

    IInstrumentProxy * i = m_InstrumentList->current ();
    if (pig && m_HasGLContext && i) {
        // Apply the color.
        i->addMass (pig, amount);
    }
    if (colorChangedByMainWindow) {
        colorChangedByMainWindow = false;

        // Update the color indicator.
        if (pig->isValid ()) {
            ui->colorPicker->setColor (
                pig->absorbing (),
                pig->scattering ()
            );
        }
    } else {
        // Check if the instrument has color.
        if (i->currentAbsorbing () &&
            i->currentScattering ()) {
            // And update the color indicator.
            ui->colorPicker->setColor (
                i->currentAbsorbing (),
                i->currentScattering ()
            );
        }
    }
}

void MainWindow::userStartedPainting (void) {
    // Pick-up the color when the stroke starts?
    if (ui->colorPicker)
        ui->colorPicker->pickUp ();
}

void MainWindow::userStoppedPainting (void) {
    resetDragStates ();

    colorChangedByMainWindow = TRUE;

    if (m_InstrumentList && m_HasGLContext) {
        IInstrumentProxy * i = m_InstrumentList->current ();
        if (i) {
            if (ui->colorPicker) {
                if (m_RestoreInstrumentColor) {
                    // Set the instrument color again.
                    i->addMass (ui->colorPicker->colorExt (), 1.0f);
                } else
                    // Check if the instrument has color.
                    if (i->currentAbsorbing () &&
                        i->currentScattering ()) {
                        // And set it to colorpicker.
                        ui->colorPicker->setColor (
                            i->currentAbsorbing (),
                            i->currentScattering ()
                        );
                    }
            }
        }
    }

    //colorChangedByMainWindow = false;
    updateTitle ();
}

void MainWindow::strokeDirectionChanged (void) {
    m_UpdateBeep = TRUE;
}

void MainWindow::instrumentChanged (IInstrumentProxy * instrument) {
    if (!instrument)
        return;

    if (instrument->promoteColorChooser () == COLORMETHOD_KMSAMPLE) {
        m_Palette->askToHide ();
        m_dynPalette->askToHide();
    } else
    if (instrument->promoteColorChooser () == COLORMETHOD_PALETTE) {
        m_Palette->asWidget ()->show ();
        m_dynPalette->asWidget()->show();
    }

    /*
     * Notify the PaintingStyle widget, if this instrument
     * is not a tool-instrument.
     */
    if (instrument->instrumentGroup() != INSTRUMENTGROUP_TOOLBOX) {
        ui->paintingStyle->setBrushGroup (
            instrument->instrumentGroup() == INSTRUMENTGROUP_BRUSH
        );
    }

    /*
     * Set current color to the instrument
     * if requested.
     */
    if (!ui->checkIndependentColors->isChecked ()) {
        IPigmentMixture * pigment = ui->colorPicker->colorExt ();
        if (pigment) {
            colorChangedByColorPicker (pigment, 1.0f, COLORMETHOD_KMSAMPLE);
        }
    } else {
        if (instrument->currentAbsorbing () &&
            instrument->currentScattering ()) {
            // And update the color indicator.
            ui->colorPicker->setColor (
                instrument->currentAbsorbing (),
                instrument->currentScattering ()
            );
        }
    }
}

void MainWindow::toolInstrumentChanged (IInstrumentProxy * instrument) {
    if (instrument == NULL) {
        if (m_InstrumentList) {
            (m_InstrumentList->asWidget ())->setEnabled (TRUE);
             m_InstrumentList->pokeObject ();
        }
    } else {
        (m_InstrumentList->asWidget ())->setEnabled (FALSE);

        // Change the canvas instrument.
        if (m_Canvas)
            m_Canvas->setInstrument (instrument);
    }
}

void MainWindow::keyPressEvent (QKeyEvent * event) {
    // Check for modifiers.
    int k = event->key ();
    checkModifiers(k);
    int mk = k | event->modifiers();

    if (m_ModifierKeys.contains (mk)) {
        if (m_Canvas)
            m_Canvas->setDragMode (m_ModifierKeys.value (mk));
    } else if(m_ModifierKeysWheel.contains (mk)) {
        m_WheelCurrentModifier = m_ModifierKeysWheel.value(mk);
    } else if (m_ModifierKeys.contains (k)) {
        if (m_Canvas)
            m_Canvas->setDragMode (m_ModifierKeys.value (k));
    } else if(m_ModifierKeysWheel.contains (k)) {
        m_WheelCurrentModifier = m_ModifierKeysWheel.value(k);
    } else {
        QWidget::keyPressEvent (event);
    }
}

void MainWindow::keyReleaseEvent (QKeyEvent * event) {
    // Check for modifiers.
    int k  = event->key ();
    int mk = k | event->modifiers();

    checkModifiers(k);

    // Unset the drag-mode on any key-release event.
    if (m_Canvas)
        m_Canvas->setDragMode (DRAGSTATE_NOOP);

    if(m_ModifierKeysWheel.contains (k) || m_ModifierKeysWheel.contains(mk)) {
        m_WheelCurrentModifier = WheelDoesNothing;
		// default wheel action
		if(m_ModifierKeysWheel.contains (0)) {
			m_WheelCurrentModifier = m_ModifierKeysWheel.value(0);
		}
    } else {
        QWidget::keyReleaseEvent (event);
    }
}

void MainWindow::resetDragStates (void) {
    if (m_Canvas) {
        m_Canvas->setDragMode (DRAGSTATE_NOOP);
    }
}

void MainWindow::reloadShaders (void) {
    if (m_Connector)
        m_Connector->reloadShaders ();
}

void MainWindow::clearCanvas (void) {
    if (m_Canvas) {
        int ret = QMessageBox::warning (this, m_APP_TITLE,
                      tr ("Are you sure you want to clear the canvas and abandon all its content?"),
                      QMessageBox::Yes | QMessageBox::Default,
                      QMessageBox::No  | QMessageBox::Escape
                  );

        if (ret == QMessageBox::Yes) {
            m_Canvas->erase ();
            statusBar()->showMessage (tr ("Canvas has been erased."),  STATUSBAR_MESSAGE_TIMEOUT);
        }
    }
}

void MainWindow::dryCanvas (void) {
    if (m_Canvas) {
        if (m_Canvas->dry ()) {
            statusBar()->showMessage (tr ("Canvas has been dryed."),  STATUSBAR_MESSAGE_TIMEOUT);
        } else {
            statusBar()->showMessage (tr ("Unable to dry the canvas."),  STATUSBAR_MESSAGE_TIMEOUT);
        }
    }
}

void MainWindow::notAssigned (void) {
    QMessageBox::information (this, m_APP_TITLE, tr("Not implemented yet:\nSignal is not assigned"));
}

/// Some dock has changed it's visibility,
/// and we might want to update some events here.
void MainWindow::dockVisibility (bool visible) {
    Q_UNUSED (visible);

  //ui->actionShow_instrument_list->setEnabled   (!ui->dockInstruments->isVisible ());
    ui->actionShow_color_picker_2->setEnabled    (!ui->dockPicker->isVisible ());
    ui->actionShow_canvas_options->setEnabled    (!ui->dockCanvas->isVisible ());
}

void MainWindow::newFile (void) {
    bool isStarted = isSimulationStarted ();
    if (isStarted)
        stopSimulation ();

    if (!QGLFormat::hasOpenGL () ||
        !m_HasGLContext) {
        QMessageBox::warning (
            this,
            tr ("Qaquarelle internal error."),
            tr ("I'm sorry, but it seems that OpenGL is not available on your system, so I cannot proceed.")
        );
    } else
    if (maybeSave ()) {
        closeWorkspace ();
        setCurrentFile ("");

        PaintingSettings set;
        NewPainting np(&set, this);
        if(!np.exec()) return;

        // Create a new workspace.
        IWorkspace * workspace = g_fac_SPAWN_WORKSPACE (m_Canvas, m_Helper,
                                                        set.width, set.height, set.dpi);
        if (workspace) {
            workspace->attachLogger  (m_Logger);
            workspace->enableLogging (TRUE);
            bindWorkspace (workspace);

            // Clear our brand-new canvas now.
            m_Canvas->erase ();

            // Init the painting technique.
            switch (set.technique) {
                case NEWPAINTING_WATERCOLORS:
                    ui->paintingStyle->sync (PAINTINGSTYLE_WATERCOLOURS);
                    m_PaperSwitch->setByClass (sPaper::WATERCOLOR);
                    break;
                case NEWPAINTING_OILS:
                    ui->paintingStyle->sync (PAINTINGSTYLE_OIL);
                    m_PaperSwitch->setByClass (sPaper::CANVAS);
                    break;
                case NEWPAINTING_SKETCH:
                    ui->paintingStyle->sync (PAINTINGSTYLE_WATERCOLOURS);
                    m_PaperSwitch->setByClass (sPaper::PLAIN);
                    break;
                default:
                    ui->paintingStyle->sync (PAINTINGSTYLE_WATERCOLOURS);
                    m_PaperSwitch->setByClass (sPaper::PLAIN);
                    break;
            }

            updateTitle ();
        }
    }

    if (isStarted)
        startSimulation ();

    if (isSimulationStarted ()) {
        ui->actionStart_Simulation->setEnabled (FALSE);
        ui->actionStop_Simulation->setEnabled  (TRUE);
    } else {
        ui->actionStart_Simulation->setEnabled (TRUE);
        ui->actionStop_Simulation->setEnabled  (FALSE);
    }

    userStoppedPainting ();
    if (m_InstrumentList) {
        if (m_InstrumentList->current ()) {}
            instrumentChanged (m_InstrumentList->current ());
    }
}

void MainWindow::open (void) {
    bool isStarted = isSimulationStarted ();
    if (isStarted)
        stopSimulation ();

    if (!QGLFormat::hasOpenGL () ||
        !m_HasGLContext) {
        QMessageBox::warning (
            this,
            tr ("Qaquarelle internal error."),
            tr ("I'm sorry, but it seems that OpenGL is not available on your system, so I cannot proceed.")
        );
    } else
    if (maybeSave ()) {
        QString fileName = QFileDialog::getOpenFileName (
            this,
            QString (),
            QString (),
            tr ("QAquarelle files (*.qaq)")
        );

        if (!fileName.isEmpty ()) {
            // Load a binary data.
            if (fileName.endsWith (".qaq",  Qt::CaseInsensitive)) {
                closeWorkspace ();
                if (loadFile (fileName)) {
                    setCurrentFile (fileName);
                } else {
                    QMessageBox::warning (
                        this,
                        tr ("QAquarelle loading error."),
                        tr ("I'm sorry, but I'm unable to parse this file. Most probably its format was deprecated.")
                    );
                }
            } else {
                QMessageBox::warning (
                    this,
                    tr ("QAquarelle loading error."),
                    tr ("You have specified the file type that is unknown to me... I'm sorry.")
                );
            }
        }
    }

    if (isStarted)
        startSimulation ();

    if (isSimulationStarted ()) {
        ui->actionStart_Simulation->setEnabled (FALSE);
        ui->actionStop_Simulation->setEnabled  (TRUE);
    } else {
        ui->actionStart_Simulation->setEnabled (TRUE);
        ui->actionStop_Simulation->setEnabled  (FALSE);
    }
}

bool MainWindow::maybeSave (void) {
    bool isStarted = isSimulationStarted ();
    if (isStarted)
        stopSimulation ();

    bool result = TRUE;
    if (m_Workspace) {
        if (m_Canvas->changesCheck ()) {
            int ret = QMessageBox::warning (this, m_APP_TITLE,
                        tr("Something has been modified in the workspace.\n"
                           "Would you like to save your changes?"),
                        QMessageBox::Yes | QMessageBox::Default,
                        QMessageBox::No,
                        QMessageBox::Cancel | QMessageBox::Escape);

            if (ret == QMessageBox::Yes)
                result = save ();
            else if (ret == QMessageBox::Cancel)
                result = FALSE;
        }
    }

    if (isStarted)
        startSimulation ();

    return result;
}

bool MainWindow::save (void) {
    bool isStarted = isSimulationStarted ();
    if (isStarted)
        stopSimulation ();

    bool result = FALSE;
    if (m_Workspace) {
        if (m_curFile.isEmpty ()) {
            result = saveAs ();
        } else {
            result = saveFile (m_curFile);
        }
    }

    if (isStarted)
        startSimulation ();

    if (result) {
        if (m_Canvas)
            m_Canvas->changesCommit ();

        updateTitle ();
    }
    return result;
}

bool MainWindow::saveAs (void) {
    bool isStarted = isSimulationStarted ();
    if (isStarted)
        stopSimulation ();

    bool result = FALSE;
    if (m_Workspace) {
        // FIXME Ask before overwriting.
        QString fileName = QFileDialog::getSaveFileName (
            this,
            QString (),                                                                     // Caption
            QString (),                                                                     // Dir
            tr ("QAquarelle files (*.qaq);;Common image files (*.jpg *.png *.bmp *.gif)"),  // Filter
            NULL,                                                                           // Selected filter
            0                                                                               // Options
        );
        if (fileName.isEmpty ())
            result = FALSE;
        else {
            // Auto-append extension
            if (QFileInfo (fileName).suffix ().isEmpty ()) {
                fileName += ".qaq";
            }

            result = saveFile (fileName);
            if (result) {
                setCurrentFile (fileName);
                updateTitle ();
            }
        }
    }

    if (isStarted)
        startSimulation ();

    return result;
}

void MainWindow::clean (void) {
    if (maybeSave ()) {
        closeWorkspace ();
    }
}

void MainWindow::about (void) {
      QMessageBox::about (this, tr ("About QAquarelle"),
            tr ("<b>QAquarelle</b> is the open-source graphical simulator.<br />"
               "Its goal is to provide a natural painter's environment.<br />"
               "You're welcome to visit <a href=\"http://qaquarelle.sourceforge.net\">"
               "the QAquarelle website</a> for more information."));
}

void MainWindow::setCurrentFile (const QString &fileName) {
    m_curFile = fileName;

    updateTitle ();
}

QString strippedName (const QString &fullFileName) {
    return QFileInfo (fullFileName).fileName ();
}

void MainWindow::updateTitle (void) {
    if (m_Workspace && m_Canvas) {
        // If we have workspace loaded
        // show the info about the canvas
        QString shownName;
        if (m_curFile.isEmpty ())
            shownName = "untitled";
        else
            shownName = strippedName (m_curFile);

        if (m_Canvas->changesCheck ()) {
            setWindowTitle    ("[" + m_APP_TITLE + "] " + shownName + " [*]");
            setWindowModified (TRUE);
        } else {
            setWindowTitle ("[" + m_APP_TITLE + "] " + shownName);
            setWindowModified (FALSE);
        }
    } else {
        // If nothing is loaded
        // then just show the title
        setWindowTitle    (m_APP_TITLE);
        setWindowModified (FALSE);
    }
}

void MainWindow::closeEvent (QCloseEvent *event) {
    if (maybeSave ()) {
        closeWorkspace ();
        event->accept();
    } else {
        event->ignore();
    }
}

void MainWindow::closeSubWindow(bool* allow)
{
    if (maybeSave ()) {
        closeWorkspace ();
        setCurrentFile ("");
		*allow = true;
	} else {
		*allow = false;
	}
}

void MainWindow::preferences()
{
    const int last_texcompression = PREF.ogl_textureCompression;

    PreferencesDialog pd;
    pd.exec();
    if(m_Canvas) {
        m_Canvas->setSmartStrokeEnabled(PREF.smartStroke);
    }

    ui->colorPicker->setPickerStyle(PREF.colorPickerStyle);
	if(m_Canvas) {
	  if(PREF.pressureControl)
		m_Canvas->setPressureWorkingRange(PREF.pressureMin, PREF.pressureMax);
	  else
		m_Canvas->setPressureWorkingRange(0.f, 1.f);
	}

	createHotkeys();

    // Apply the texture compression if changed.
    if (last_texcompression != PREF.ogl_textureCompression) {
        // Force the canvas to recreate all
        // its OpenGL-related objects.
        if (m_Canvas) {
            for (int i = 0; ; i++) {
                IArrayGridPatch * patch = m_Canvas->patchIterator (i);
                if (!patch)
                    break;

                patch->data ()->purgeToSW ();
            }

            // And update the settings.
            if (m_Helper)
                m_Helper->setPerformanceParameter (OPENGLHELPER_TEX_COMPRESSION,
                                                   PREF.ogl_textureCompression);
            statusBar()->showMessage (tr ("Changed the texture compression."),
                                      STATUSBAR_MESSAGE_TIMEOUT);
        }
    }
}

void MainWindow::paintingStyleChanged (void) {
    if (m_InstrumentList && m_HasGLContext) {
        IInstrumentProxy * i = m_InstrumentList->current ();
        if (i) {
            // Apply the color.
            i->addMass (ui->colorPicker->colorExt (), 1.0f);
        }

        if (m_Workspace) {
            // Change the instrument wetness attribute if supported.
            const float wetness = MAX (0.00f, 0.75f - m_Workspace->paintingStyle ()->paintViscosity ());
            i->customParameterSet ("wetness", &wetness);    // Set the parameter.
            i->setPressure (0.0f, INPUT_CONTROLLER_GOD);    // Force the brush update.

            // Dry the canvas
            // (we'll change the rendering parameters for the wet layer).
            if (m_Canvas)
                m_Canvas->dry ();
        }
    }
}

void MainWindow::changeCanvasScale (double scale) {
    if (m_Canvas) {
        m_Canvas->setUniformScale (scale);
    }
}

void MainWindow::wheelEvent ( QWheelEvent * event )
{
	float scale = 1.f - event->delta() * 0.001f;
	switch(m_WheelCurrentModifier) {
	case WheelScalesBrush:
		if(m_InstrumentList->current()) {
			m_InstrumentList->current()->setScaleModifier(scale * m_InstrumentList->current()->scaleModifier());
		}
		break;
	case WheelScalesCanvas:
		if(m_Canvas) {
			m_Canvas->multiplyScale (scale);
			ui->spinScale->setValue (m_Canvas->getUniformScale ());
		}
		break;
	case WheelDoesNothing:
	default:
		;
	}
}

void MainWindow::checkModifiers(int & key)
{
	switch (key) {
	case 0: break;
	case Qt::Key_Shift: key = Qt::ShiftModifier; return;
	case Qt::Key_Control: key = Qt::ControlModifier; return;
	case Qt::Key_Alt: key = Qt::AltModifier; return;
	case Qt::Key_Meta: key = Qt::MetaModifier; return;
	}
}

void MainWindow::consciousness (void) {
    if (m_UpdateBeep) {
        m_UpdateBeep = FALSE;
        ui->iconBeep0->iconChanged (m_Skin->pixmapByName ("icons/bullet_cyan8.png"));
    } else {
        ui->iconBeep0->iconChanged (NULL);
    }
}

void MainWindow::canvasPreview (QImage * image) {
    if (image) {
    }
}

void MainWindow::backgroundTask (QString name, int percent, bool blocker) {
	PROFILE_FUNC();
    int nblockers = 0;
    for (int i = 0; i < m_ThreadTasks.size (); i++) {
        if (m_ThreadTasks[i]->blocker)
            nblockers++;
    }

    if (percent >= 100) {
        // Task is complited.
        for (int i = 0; i < m_ThreadTasks.size (); i++) {
            sBgTask * t = m_ThreadTasks[i];
            if (t->name == name) {
                m_ThreadTasks.removeAt (i);
                delete t;

                // Unlock the canvas when there are
                // no blockers left.
                if (nblockers == 1 && m_Canvas) {
                    m_Canvas->asWidget ()->setEnabled (TRUE);
                    setEnabled (TRUE);
                }

                return;
            }
        }
    } else {
        sBgTask * t = NULL;
        for (int i = 0; i < m_ThreadTasks.size (); i++) {
            if (m_ThreadTasks[i]->name == name) {
                t = m_ThreadTasks[i];
                break;
            }
        }

        // Add a new task if bot found.
        if (!t) {
            t = new sBgTask;
            t->name    = name;
            t->blocker = blocker;

            m_ThreadTasks.append (t);

            // Lock the canvas if this task
            // is a blocker.
            if (blocker && m_Canvas) {
                m_Canvas->asWidget ()->setEnabled (FALSE);
                setEnabled (FALSE);
            }
        }

        if (t) {
            // Noop yet.
        }
    }
}

void MainWindow::renderingOptsChanged (void) {
    if (ui->actionPaint_bump_mapping->isChecked ()) {
        if (!m_Block)
            PREF.render_paintBump = TRUE;
        if (m_Canvas)
            m_Canvas->enableRenderingOptions (RENDER_PAINT_BUMP);
    } else {
        if (!m_Block)
            PREF.render_paintBump = FALSE;
        if (m_Canvas)
            m_Canvas->disableRenderingOptions (RENDER_PAINT_BUMP);
    }

    if (ui->actionPaper_bump_mapping->isChecked ()) {
        if (!m_Block)
            PREF.render_paperBump = TRUE;
        if (m_Canvas)
            m_Canvas->enableRenderingOptions (RENDER_PAPER_BUMP);
    } else {
        if (!m_Block)
            PREF.render_paperBump = FALSE;
        if (m_Canvas)
            m_Canvas->disableRenderingOptions (RENDER_PAPER_BUMP);
    }

    if (ui->actionMotion_vectors->isChecked ()) {
        if (!m_Block)
            PREF.render_mvectors = TRUE;
        if (m_Canvas)
            m_Canvas->enableRenderingOptions (RENDER_MOTION_VECTORS);
    } else {
        if (!m_Block)
            PREF.render_mvectors = FALSE;
        if (m_Canvas)
            m_Canvas->disableRenderingOptions (RENDER_MOTION_VECTORS);
    }

    if (m_Canvas)
        m_Canvas->asWidget()->update ();
}

void MainWindow::renderingOptsRequested (void) {
    // Apply the rendering options.
    m_Block = TRUE;
    renderingOptsChanged ();
    m_Block = FALSE;
}
