/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "arrays_gpu.h"
#if !defined(__arm__) && !defined(__aarch64__)
#include <mm_malloc.h>
#else
#include <stdlib.h>
#define _mm_malloc(size, alignment) malloc(size)
#define _mm_free(ptr) free(ptr)
#endif
#include <memory.h>
#include <assert.h>


//#ifdef QT_OPENGL_LIB

CArray_gpu::CArray_gpu (IOpenGLHelper * helper,
                        LONG width, LONG height,
                        GLenum format, GLenum type,
                        GLuint src,
                        int flags) {
    locked        = NULL;
    lockedu       = NULL;
    mapped        = NULL;
    dataformat    = format;
    transfer_type = GL_LUMINANCE;
    transfer_prec = type;
    block_elsz    = 1;
    modified      = FALSE;
    fbo_is_shared = FALSE;

    m_Flags       = flags;
    m_Id          = ISpawnedObject::allocateId ();
    m_Helper      = helper;
    if (m_Helper)
        m_Helper->addRef ();

    ISpawnedObject::_register (this);

    switch (type) {
        case GL_FLOAT:
        case GL_UNSIGNED_INT:
        case GL_INT:
            block_elsz = 4;
            break;
        case GL_UNSIGNED_SHORT:
        case GL_SHORT:
            block_elsz = 2;
            break;
    }

    switch (dataformat) {
        case GL_RGBA:
        case GL_RGBA8:
        case GL_RGBA16F_ARB:
        case GL_RGBA32F_ARB:
        case GL_COMPRESSED_RGBA_ARB:
            transfer_type = GL_RGBA;
            block_elsz   *= 4;
            break;
        case GL_RGB8:
        case GL_COMPRESSED_RGB_ARB:
            transfer_type = GL_RGB;
            block_elsz   *= 3;
            break;
        case GL_LUMINANCE8:
        case GL_LUMINANCE16F_ARB:
        case GL_LUMINANCE32F_ARB:
        case GL_INTENSITY:
        case GL_INTENSITY8:
        case GL_COMPRESSED_LUMINANCE_ARB:
            transfer_type = GL_LUMINANCE;
            break;
    }

    block_sz   = width * height * block_elsz;
    block_w    = width;
    block_h    = height;

    state  = ARRAY_STATE__EMPTY;
    tex    = UINT_UNDEFINED;
    fbo    = UINT_UNDEFINED;
    fbo_w  = NULL;

    buf_usage = GL_DYNAMIC_DRAW;

#ifdef ENABLE_PBO_TRANSFER
    allow_pbo = GLEW_VERSION_2_0 && m_Helper->checkExtension ("GL_ARB_pixel_buffer_object");
    buf       = UINT_UNDEFINED;
#endif

    if (src != UINT_UNDEFINED) {
        // Import the data if you are provided
        // with the source texture.
        import_tex (src);
    } else {
        // Or fill with zeros if there's
        // no data available.
#ifdef ARRAYS_FILL_WITH_ZEROS
        zero ();
#endif
    }
}

CArray_gpu::~CArray_gpu (void) {
    ISpawnedObject::_unregister (this);

    if (NULL == fbo_w) {
        if (fbo != UINT_UNDEFINED && !fbo_is_shared)
            glDeleteFramebuffersEXT (1, &fbo);
    }

    if (tex != UINT_UNDEFINED)
        glDeleteTextures (1, &tex);

#ifdef ENABLE_PBO_TRANSFER
    if (buf != UINT_UNDEFINED)
        glDeleteBuffers (1, &buf);
#endif

    if (locked)
        _mm_free (locked);

    //SAFE_FREE    (lockedu);
    SAFE_RELEASE (m_Helper);
}

void CArray_gpu::purgeToSW (void) {
    lock_pointer (GL_READ_WRITE_ARB);
    unlock ();

    if (NULL == fbo_w) {
        if (fbo != UINT_UNDEFINED && !fbo_is_shared) {
            glDeleteFramebuffersEXT (1, &fbo);
            fbo = UINT_UNDEFINED;
        }
    } else {
        fbo_w = NULL;
        fbo   = UINT_UNDEFINED;
    }

    if (tex != UINT_UNDEFINED) {
        glDeleteTextures (1, &tex);
        tex = UINT_UNDEFINED;
    }

#ifdef ENABLE_PBO_TRANSFER
    if (buf != UINT_UNDEFINED) {
        glDeleteBuffers (1, &buf);
        buf = UINT_UNDEFINED;
    }
#endif
}

void CArray_gpu::obtain_FBO (void) {
    if (fbo == UINT_UNDEFINED) {
        if (block_w == m_Helper->sharedPatchSize () &&
            block_h == m_Helper->sharedPatchSize () &&
            !(m_Flags & FLAGS_NO_SHARE)) {
            // If our dimentions are the same as
            // the promoted patch dimentions we
            // shall use a shared FBO.
            fbo_w = m_Helper->sharedFBO (OPENGLHELPER_FBO_PATCH);
            if (fbo_w) {
                fbo = fbo_w->handle ();
                fbo_is_shared = TRUE;
            }
        } else {
            glGenFramebuffersEXT (1, &fbo);
            fbo_is_shared = FALSE;
            fbo_w         = NULL;
        }
    }
}

void CArray_gpu::obtain_PBO (void) {
#ifdef ENABLE_PBO_TRANSFER
    if (buf == UINT_UNDEFINED) {
        if (block_w == m_Helper->sharedPatchSize () &&
            block_h == m_Helper->sharedPatchSize () &&
            FALSE) {
            // If our dimentions are the same as
            // a promoted patch dimentions we
            // might use a shared PBO object.. but we won't.
            // FIXME Shared PBO is not implemented.
            // ...
        } else {
            glGenBuffersARB (1, &buf);
            glBindBufferARB (GL_PIXEL_PACK_BUFFER_ARB, buf);
            glBufferDataARB (GL_PIXEL_PACK_BUFFER_ARB, block_sz, NULL, buf_usage);
            glBindBufferARB (GL_PIXEL_PACK_BUFFER_ARB, 0);
        }
    }
#endif
}

void CArray_gpu::import_tex (GLuint tex) {
    GLuint target = lock_gltex ();
    if (target != UINT_UNDEFINED) {
        obtain_FBO ();

        m_Helper->bindTexture_ex (target, 0);

        // Attach the source texture to FBO.
        if (fbo_w) {
            fbo_w->bind  ();
            glReadBuffer (fbo_w->attachTexture (tex));
        } else {
            glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, fbo);
            glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
            glReadBuffer (GL_COLOR_ATTACHMENT0_EXT);
        }

        // And duplicate it.
        glCopyTexSubImage2D  (GL_TEXTURE_2D, 0, 0, 0, 0, 0, block_w, block_h);
        unlock ();
    }
}

/// Performs lazy downloading of texture form GPU.
void * CArray_gpu::lock_pointer (int x, int y, int w, int h, GLenum access) {
    PROFILE_FUNC();

    if (state == ARRAY_STATE__HERE_LOCKED ||
        state == ARRAY_STATE__THERE_LOCKED) {
        // This should never happen ><'...
        return NULL;
    }

    if (w > block_w ||
        h > block_h) {
        return NULL;
    }

    BOOL recreated = FALSE;

#ifdef ENABLE_PBO_TRANSFER
    // Check if we're able to use pbo now.
    if (allow_pbo && m_Helper->inited ()) {
        buf_access = access;

        // Bind our PBO.
        obtain_PBO ();
        glBindBufferARB (GL_PIXEL_PACK_BUFFER_ARB, buf);

        // Read the data back if needed.
        if (buf_access != GL_WRITE_ONLY &&
            state == ARRAY_STATE__THERE &&
            tex   != UINT_UNDEFINED) {

            obtain_FBO ();
            if (fbo_w) {
                fbo_w->bind ();
                glReadBuffer (fbo_w->attachTexture (tex));
            } else {
                glBindFramebufferEXT      (GL_FRAMEBUFFER_EXT, fbo);
                glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
                glReadBuffer              (GL_COLOR_ATTACHMENT0_EXT);
            }

            // Read our data back.
            if (w && h) {
                glReadPixels (x, y, w, h, transfer_type, transfer_prec, BUFFER_OFFSET(0));
            } else {
                glReadPixels (0, 0, block_w, block_h, transfer_type, transfer_prec, BUFFER_OFFSET(0));
            }
        }

        // Get the pointer by mapping the buffer.
        mapped = (BYTE*) glMapBuffer (GL_PIXEL_PACK_BUFFER_ARB, access);
        locked = NULL;
        SAFE_FREE (lockedu);

        glBindBuffer (GL_PIXEL_PACK_BUFFER_ARB, 0);
    } else {
#endif
    mapped = NULL;
    if (!locked) {
        // FIXME Share the sw buffer to avoid often reallocations?
        //lockedu = (BYTE*) malloc (block_sz + 64);
        //locked  = (BYTE*) ALIGN32 (lockedu);

        locked    = (BYTE*) _mm_malloc (block_sz + 64, 32);
        recreated = TRUE;
    }

    if (state == ARRAY_STATE__THERE && tex != UINT_UNDEFINED) {
        m_Helper->bindTexture_ex (tex, 0);

        obtain_FBO ();
        if (fbo_w) {
            fbo_w->bind ();
            glReadBuffer (fbo_w->attachTexture (tex));
        } else {
            glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, fbo);
            glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
            glReadBuffer (GL_COLOR_ATTACHMENT0_EXT);
        }

        glDrawBuffer (GL_NONE);

        // Read all data from texture.
        if (w && h) {
            glReadPixels (x, y, w, h, transfer_type, transfer_prec, locked);
        } else {
            glReadPixels (0, 0, block_w, block_h, transfer_type, transfer_prec, locked);
        }

        glReadBuffer (GL_NONE);
        //glGetTexImage (GL_TEXTURE_2D, 0, transfer_type, transfer_prec, locked);
    }

#ifdef ENABLE_PBO_TRANSFER
    }
#endif

    modified = TRUE;
    if (mapped) {
        state = ARRAY_STATE__HERE_LOCKED;
        return (void*) mapped;
    } else
    if (locked) {
        state = ARRAY_STATE__HERE_LOCKED;
        return (void*) locked;
    } else
        return NULL;
}

void CArray_gpu::unlock (void) {
    switch (state) {
        case ARRAY_STATE__HERE_LOCKED:
#ifdef ENABLE_PBO_TRANSFER
            if (mapped && allow_pbo) {
                if (buf != UINT_UNDEFINED) {
                    obtain_PBO ();
                    glBindBufferARB  (GL_PIXEL_PACK_BUFFER_ARB, buf);
                    glUnmapBufferARB (GL_PIXEL_PACK_BUFFER_ARB);
                    glBindBufferARB  (GL_PIXEL_PACK_BUFFER_ARB, 0);
                }

                mapped = NULL;
            }
#endif
            state = ARRAY_STATE__HERE;
            break;
        case ARRAY_STATE__THERE_LOCKED:
            state = ARRAY_STATE__THERE;
            break;
        default:
            break;
    }
}

void CArray_gpu::fill (BYTE r, BYTE g, BYTE b, BYTE a) {
    PROFILE_FUNC();

    assert (state != ARRAY_STATE__THERE_LOCKED);
    assert (state != ARRAY_STATE__HERE_LOCKED);

    if (state == ARRAY_STATE__THERE) {
        fill (
            r / 255.0f,
            g / 255.0f,
            b / 255.0f,
            a / 255.0f
        );
    } else {
        BYTE * p = (BYTE*) lock_pointer (GL_WRITE_ONLY_ARB);
        if (p) {
            BYTE target[16] __attribute__((aligned(16))) = {
                r, g, b, a,
                r, g, b, a,
                r, g, b, a,
                r, g, b, a
            };

            for (int i = block_sz / 16; i; i--) {
                memcpy (p, target, 16);
                p += 16;
            }

            if (block_sz % 16)
                memcpy (p, target, block_sz % 16);

            modified = TRUE;
            unlock ();
        }
    }
}

void CArray_gpu::fill (float r, float g, float b, float a) {
    assert (state != ARRAY_STATE__THERE_LOCKED);
    assert (state != ARRAY_STATE__HERE_LOCKED);

    obtain_FBO ();
    if (lock_gltex ()) {
        if (fbo_w) {
            fbo_w->bind ();
            glDrawBuffer (fbo_w->attachTexture (tex));
        } else {
            glBindFramebufferEXT      (GL_FRAMEBUFFER_EXT, fbo);
            glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
            glDrawBuffer              (GL_COLOR_ATTACHMENT0_EXT);
        }

        glClearColor (r, g, b, a);
        glClear      (GL_COLOR_BUFFER_BIT);
        unlock ();
    }
}

IArray * CArray_gpu::duplicate (void) {
    if (state == ARRAY_STATE__HERE_LOCKED ||
        state == ARRAY_STATE__THERE_LOCKED) {

        assert (state != ARRAY_STATE__THERE_LOCKED);
        assert (state != ARRAY_STATE__HERE_LOCKED);

        // This should never happen... ><'
        return NULL;
    }

    if (state == ARRAY_STATE__HERE) {
        IArray * replica = new CArray_gpu (m_Helper, block_w, block_h, dataformat, transfer_prec,
                                           UINT_UNDEFINED, m_Flags);
        if (replica) {
                  BYTE * dst = (BYTE*) replica->lock_pointer ();
            const BYTE * src = (const BYTE*) lock_pointer_const ();

            // Duplicate in RAM.
            memcpy (dst, src, block_sz);
            replica->unlock ();
            unlock ();

            return replica;
        }
    } else
    if (state == ARRAY_STATE__THERE) {
        GLuint tex = lock_gltex ();
        unlock ();
        if (tex != UINT_UNDEFINED) {
            // Duplicate in GPU.
            return new CArray_gpu (m_Helper, block_w, block_h, dataformat, transfer_prec,
                                   tex, IArray::FLAGS_NULL);
        }
    }

    return NULL;
}

/// Performs the lazy texture downloading into GPU.
LONG CArray_gpu::lock_gltex (void) {
    PROFILE_FUNC();

    if (state == ARRAY_STATE__HERE_LOCKED ||
        state == ARRAY_STATE__THERE_LOCKED) {

        assert (state != ARRAY_STATE__THERE_LOCKED);
        assert (state != ARRAY_STATE__HERE_LOCKED);

        // This should never happen... ><'
        return UINT_UNDEFINED;
    }

    BOOL recreated = FALSE;

    // Create a texture if
    // haven't created yet
    if (tex == UINT_UNDEFINED) {
        glGenTextures (1, &tex);
        if (tex != UINT_UNDEFINED) {
            m_Helper->bindTexture_ex (tex, 0);

            GLint compressed;
            GLint internalformat;

            // Set the texture parameters.
            glTexImage2D (GL_TEXTURE_2D, 0, dataformat, block_w, block_h, 0, transfer_type, transfer_prec, 0);

            // Check if the image has been compressed by GL.
            glGetTexLevelParameteriv (GL_TEXTURE_2D, 0,
                                      GL_TEXTURE_COMPRESSED_ARB, &compressed);

            // Query for the internal format.
            glGetTexLevelParameteriv (GL_TEXTURE_2D, 0,
                                      GL_TEXTURE_INTERNAL_FORMAT, &internalformat);

            m_Helper->textureCreationCallback (
                dataformat,
                internalformat,
                compressed
            );

            recreated = TRUE;
        } else {
            // This should never happen... ><'
            assert (m_Helper->checkGLErrors () == NULL);
        }
    } else {
        m_Helper->bindTexture_ex (tex, 0);
    }

    // Pack our data back to hardware
    if (state == ARRAY_STATE__HERE) {
        if (modified || recreated) {
#ifdef ENABLE_PBO_TRANSFER
            if (allow_pbo) {
                obtain_PBO ();
                glBindBufferARB (GL_PIXEL_UNPACK_BUFFER_ARB, buf);
                glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, block_w, block_h, transfer_type, transfer_prec, BUFFER_OFFSET(0));
                glBindBufferARB (GL_PIXEL_UNPACK_BUFFER_ARB, 0);
            } else
#endif
            if (locked) {
                if (m_Flags & FLAGS_MIPMAPPING) {
                    m_Helper->bindTexture_ex (tex, 0, GL_LINEAR_MIPMAP_LINEAR);
                }

                // Transfer all data to texture.
                if (m_Helper->gpuVendor () == GPUVENDOR_ATI) {
                    glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, block_w, block_h, transfer_type, transfer_prec, locked);
                } else {
                    glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, block_w, block_h, transfer_type, transfer_prec, locked);
                }

                // Recreate the mipmaps if needed.
                if (m_Flags & FLAGS_MIPMAPPING) {
                    glGenerateMipmapEXT (GL_TEXTURE_2D);
                }
            }
        }
    }

    modified = FALSE;
    if (tex != UINT_UNDEFINED) {
        state = ARRAY_STATE__THERE_LOCKED;
    }

    return tex;
}

LONG CArray_gpu::swap (LONG new_handle) {
    DWORD ret = lock_gltex ();
    if (ret != UINT_UNDEFINED) {
        tex = new_handle;

        // Delete the RAM-copy if any
        if (locked)
            _mm_free (locked);

        //SAFE_FREE (lockedu);
        locked  = NULL;
        lockedu = NULL;
        unlock ();

        return ret;
    } else
        return new_handle;
}

IArray * g_fac_SPAWN_ARRAY (IOpenGLHelper * helper, LONG width, LONG height,
                            GLenum format, GLenum type, int flags) {
    return new CArray_gpu (helper, width, height, format, type, UINT_UNDEFINED, flags);
}

//#endif
