/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "tool.h"

ToolInstrument::ToolInstrument (int type,
                                IInstrumentFactory * factory,
                                IOpenGLHelper      * helper,
                                ILogger            * logger) {
    m_Factory = factory;
    m_Factory->addRef ();

    m_Logger = logger;
    if (m_Logger)
        m_Logger->addRef ();

    m_Helper = helper;
    if (m_Helper)
        m_Helper->addRef ();

    m_Type = type;
    m_Id   = ISpawnedObject::allocateId ();

    m_HoverPigment = g_fac_SPAWN_PIGMENT_DESCRIPTION ();

    ISpawnedObject::_register (this);

    m_Pressure       = 0.0f;
    m_Stroke         = FALSE;
    m_PositionCached = FALSE;
    m_Selected       = FALSE;

    m_LastX = -1;
    m_LastY = -1;

    m_Image.load (":/res/pencil.png");
}

ToolInstrument::~ToolInstrument (void) {
    ISpawnedObject::_unregister (this);

    SAFE_RELEASE (m_HoverPigment);
    SAFE_RELEASE (m_Factory);
    SAFE_RELEASE (m_Helper);
    SAFE_RELEASE (m_Logger);
}

void ToolInstrument::applyInstrument (IChannel * channel, ICanvasDelegate * delegate) {
    if (channel && delegate &&
        m_LastX >= 0 &&
        m_LastY >= 0) {

        if (channel->uid () != CHANNELUID_Optical)
            return;

        if (m_Type == TYPE_SELECTION) {
            // Broadcast the new selection.
            if (m_Selected)
                delegate->setSelection (
                    MIN (m_Selection_hand.x (),  m_Selection_pivot.x ()),
                    MIN (m_Selection_hand.y (),  m_Selection_pivot.y ()),
                    abs (m_Selection_hand.x () - m_Selection_pivot.x ()),
                    abs (m_Selection_hand.y () - m_Selection_pivot.y ())
                );

            return;
        }

        // Choose a cell and pick a color.
        IArrayGrid * grid = channel->lock ();
        if (grid) {
            IArrayGridPatch * cell = grid->touched (m_LastX, m_LastY);
            if (cell) {
                const BYTE * p = reinterpret_cast<const BYTE*> (cell->data ()->lock_pointer_const ());
                if (p) {
                    bool hover = FALSE;
                    const int x = m_LastX - cell->rect ()->left;
                    const int y = m_LastY - cell->rect ()->top;
                    if (x >= 0 &&
                        y >= 0) {
                        const BYTE * pa = p + 4 * (x + y * m_Helper->sharedPatchSize ());
                        const BYTE * ps = pa + m_Helper->sharedPatchSize () * 2;

                        float a[4];
                        float s[4];

                        // Check if we have a paint in wet channel.
                        if (pa[3] > 1) {
                            a[0] = pa[0] / 255.0f;
                            a[1] = pa[1] / 255.0f;
                            a[2] = pa[2] / 255.0f;
                            a[3] = pa[3] / 255.0f;
                            s[0] = ps[0] / 255.0f;
                            s[1] = ps[1] / 255.0f;
                            s[2] = ps[2] / 255.0f;
                            s[3] = ps[3] / 255.0f;

                            hover = TRUE;
                        } else {
                            // Otherwise, try to use the dry channel.
                            pa += m_Helper->sharedPatchSize () * m_Helper->sharedPatchSize () / 2;
                            ps += m_Helper->sharedPatchSize () * m_Helper->sharedPatchSize () / 2;
                        }

                        m_HoverPigment->setAbsorbing  (a);
                        m_HoverPigment->setScattering (s);
                    }

                    cell->data ()->unlock ();

                    // Broadcast the hovered pigment.
                    if (hover) {
                        delegate->pickColor (m_HoverPigment);
                    }
                }
            }

            channel->unlock ();
        }
    }
}

IInstrument * ToolInstrument::replicate (void) {
    if (m_Factory) {
        IInstrument * replica = m_Factory->create (INSTRUMENTS_IN_PACK + m_Type, m_Helper);

        return replica;
    }

    return NULL;
}

void ToolInstrument::setPosition (int x, int y, eInputMethod source) {
    Q_UNUSED (source);
    m_LastX = x;
    m_LastY = y;

    if (m_Type == TYPE_SELECTION &&
        IS_SIGNIFICANT (m_Pressure)) {
        if (m_Selected) {
            m_Selection_hand = QPoint (x, y);
        } else {
            m_Selection_pivot = QPoint (x, y);
            m_Selection_hand  = QPoint (x, y);
            m_Selected = TRUE;
        }
    }
}

void ToolInstrument::setPressure (float pressure, eInputMethod source) {
    Q_UNUSED (source);
    m_Pressure = pressure;

    if (m_Type == TYPE_SELECTION) {
        if (!IS_SIGNIFICANT (m_Pressure) && m_Selected) {
            m_Selected = FALSE;
        }
    }
}
