/***************************************************************************
 *   Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>              *
 *                                                                         *
 *   This file is a part of QAquarelle project.                            *
 *                                                                         *
 *   QAquarelle is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "brushtuft.h"
#include "qhasmath.h"
#include <memory.h>


static const float g_SMOOTH     = 0.0f;
static const float g_SMOOTH_VTX = 0.05f;


BrushTuft::BrushTuft (void) {
    m_SkelLength     = 0;
    m_Elasticity     = 0.4f;
    m_Radius         = 0.2f;

    reset ();
}

void BrushTuft::setFriction (float fx, float fy) {
    m_Friction_angle = (1.0f - g_SMOOTH) * atan2 (fy, fx) +
                               g_SMOOTH  * m_Friction_angle;
}

void BrushTuft::setGround (float y) {
    m_Ground_y = y;
}

void BrushTuft::setTilt (float tx, float ty) {
    const float xtilt = cosf (tx * 3.14f / 180.0f);
    const float ytilt = cosf (ty * 3.14f / 180.0f);

    m_Tilt_modulus = sqrtf (xtilt * xtilt + ytilt * ytilt);
    m_Tilt_angle   = (1.0f - g_SMOOTH) * atan2 (ytilt, xtilt)
                           + g_SMOOTH  * m_Tilt_angle;
}

void BrushTuft::validate (void) {
    const int n = m_SkelLength - 1;

    m_Skeleton[n].setX (MAX (0.0f, m_Ground_y - m_Skeleton[n].y ()));
    m_Skeleton[n].setZ (MAX (0.1f, 0.25f * m_Skeleton[n].x ()));
    m_Skeleton[n].setY (MIN (m_Ground_y, m_Skeleton[n].y0));

    float x[4] = {
        m_Skeleton[0].x (),
        m_Skeleton[1].x (),
        m_Skeleton[n].x (),
        0
    };

    float y[4] = {
        m_Skeleton[0].y (),
        m_Skeleton[1].y (),
        m_Skeleton[n].y (),
        0
    };


    // Estimate the node position
    // by using a B-spline.
    for (int i = 2; i < m_SkelLength - 1; i++) {
        BSpline::interpolate_quad (x, y,
            (m_Skeleton[i].y () - m_Skeleton[1].y ()) /
            (m_Skeleton[n].y () - m_Skeleton[1].y ())
        );

        m_Skeleton[i].setX (x[3]);
        m_Skeleton[i].setY (y[3]);
    }

    // Brush rotation angle.
    const float angle = (1.0f - m_Tilt_modulus) * m_Friction_angle +
                                m_Tilt_modulus  * m_Tilt_angle;

    m_TotalAngle = angle;
}

sTuftNode * BrushTuft::addNode (float x, float y, float z, bool spine, bool fixed) {
    if (m_SkelLength >= BRUSH_MAX_SKEL_LENGTH)
        return NULL;

    sTuftNode * node = m_Skeleton + m_SkelLength++;

    node->x0 = x;
    node->y0 = y;
    node->z0 = z;

    node->setX (x);
    node->setY (y);
    node->setZ (z);

    // Node is unlinked.
    node->nlinks = 0;
    node->fixed  = fixed;
    node->spine  = spine;

    return node;
}

void BrushTuft::linkNodes (sTuftNode * node0, sTuftNode * node1) {
    if (node0 && node1) {
        if (node0->nlinks < BRUSH_MAX_NODE_LINKS &&
            node1->nlinks < BRUSH_MAX_NODE_LINKS) {

            // Link.
            node0->linkage[node0->nlinks] = node1;
            node1->linkage[node1->nlinks] = node0;

            // Setup the initial edge length.
            const float length = node0->distanceTo (node1);
            node0->length0[node0->nlinks] = length;
            node0->length1[node0->nlinks] = length;
            node1->length0[node1->nlinks] = length;
            node1->length1[node1->nlinks] = length;

            // Increment.
            node0->nlinks++;
            node1->nlinks++;
        }
    }
}

float BrushTuft::lerp_vtx (float a, float b) const {
    return MATH::CLAMP_m11 (g_SMOOTH_VTX * a + (1.0f - g_SMOOTH_VTX) * b);
}

const float * BrushTuft::getGeometry (int & nVertices) {
    float * v1 = vtx_output;

    const int n = m_SkelLength - 1;
    const float cx = 0.0f;
    const float cy = 0.0f;

    const float penetration = MATH::CLAMP_01 (m_Ground_y - m_Skeleton[n].y ());

    v1[0] = cx;
    v1[1] = cy;
  //v1[2] = 0.0f;
    v1[3] = 0.5f;
    v1[4] = 0.5f;
    v1 += 7;

    // Left-top
  //v1[0] = MIN (-0.1f, cx);
  //v1[1] = MIN (-0.1f, cy - m_Skeleton[n].z () * m_Elasticity);
    v1[0] = MAX (-0.5f, cx - m_Radius - penetration * m_Elasticity);
    v1[1] = MAX (-0.5f, cy - m_Radius - penetration * m_Elasticity);
  //v1[2] = 0.0f;
    v1[3] = 0.0f;
    v1[4] = 0.0f;
    v1 += 7;

    // Right-top.
  //v1[0] = MAX (+0.1f, cx + m_Skeleton[n].x () * m_Elasticity);
  //v1[1] = MIN (-0.1f, cy - m_Skeleton[n].z () * m_Elasticity);
    v1[0] = MIN (+0.5f, cx + m_Radius + penetration * m_Elasticity);
    v1[1] = MAX (-0.5f, cy - m_Radius - penetration * m_Elasticity);
  //v1[2] = 0.0f;
    v1[3] = 1.0f;
    v1[4] = 0.0f;
    v1 += 7;

    // Right-bottom.
  //v1[0] = MAX (+0.1f, cx + m_Skeleton[n].x () * m_Elasticity);
  //v1[1] = MAX (+0.1f, cy + m_Skeleton[n].z () * m_Elasticity);
    v1[0] = MIN (+0.5f, cx + m_Radius + penetration * m_Elasticity);
    v1[1] = MIN (+0.5f, cy + m_Radius + penetration * m_Elasticity);
  //v1[2] = 0.0f;
    v1[3] = 1.0f;
    v1[4] = 1.0f;
    v1 += 7;

    // Left-bottom.
  //v1[0] = MIN (-0.1f, cx);
  //v1[1] = MAX (+0.1f, cy + m_Skeleton[n].z () * m_Elasticity);
    v1[0] = MAX (-0.5f, cx - m_Radius - penetration * m_Elasticity);
    v1[1] = MIN (+0.5f, cy + m_Radius + penetration * m_Elasticity);
  //v1[2] = 0.0f;
    v1[3] = 0.0f;
    v1[4] = 1.0f;
    v1 += 7;

    nVertices = 5;

    // Transform.
    v1 = vtx_output;
    for (int i = 0; i < nVertices; i++) {
        const float x = v1[0];
        const float y = v1[1];
        v1[0] = x * cosf (m_TotalAngle) - y * sinf (m_TotalAngle);
        v1[1] = x * sinf (m_TotalAngle) + y * cosf (m_TotalAngle);
        v1 += 7;
    }

    // Smooth.
    v1 = vtx_output;
    float * v0 = vtx_backup;
    for (int i = 0; i < nVertices; i++) {
        v1[0] = lerp_vtx (v0[0], v1[0]);
        v1[1] = lerp_vtx (v0[1], v1[1]);
        v0 += 7;
        v1 += 7;
    }

    // Set up the texcoords.
    v1 = vtx_output;
    for (int i = 0; i < nVertices; i++) {
        v1[5] = 0.5f + v1[0];
        v1[6] = 0.5f + v1[1];
        v1 += 7;
    }

    memcpy (vtx_backup, vtx_output, sizeof(float) * (nVertices * 7));
    return vtx_output;
}

void BrushTuft::reset (void) {
    // Restore the positions.
    for (int i = 0; i < m_SkelLength; i++) {
        sTuftNode * node = m_Skeleton + i;
        node->setX (node->x0);
        node->setY (node->y0);
        node->setZ (node->z0);
        for (int j = 0; j < node->nlinks; j++) {
            node->length1[j] = node->length0[j];
        }
    }

    m_Friction_angle = 0.0f;
    m_Tilt_angle     = 0.0f;
    m_Tilt_modulus   = 0.0f;
    m_TotalAngle     = 0.0f;
    m_Ground_y       = 0.0f;

    memset (vtx_backup, 0, sizeof(float) * (MAX_PRIMITIVE_VERTICES * 7));
    memset (vtx_output, 0, sizeof(float) * (MAX_PRIMITIVE_VERTICES * 7));
    float * v = vtx_backup;

    const float cx = 0.0f;
    const float cy = 0.0f;

    v[0] = cx;
    v[1] = cy;
    v += 5;

    // Left-top
    v[0] = -0.1f;
    v[1] = -0.1f;
    v += 5;

    // Right-top.
    v[0] = +0.1f;
    v[1] = -0.1f;
    v += 5;

    // Right-bottom.
    v[0] = +0.1f;
    v[1] = +0.1f;
    v += 5;

    // Left-bottom.
    v[0] = -0.1f;
    v[1] = +0.1f;
    v += 5;
}
/*
float BrushTuft::f (const float * state) const {
    float r = 0.0f;
    float bottom = 0.0f;

    for (int i = 0; i < m_SkelLength; i++) {
        // Penetration force.
        r += 0.1f * MAX (0.0f, m_Skeleton[i].y () - bottom);

        // Spring resistance.
        const float dx = m_Skeleton[i].x () - m_Skeleton[i].x0;
        const float dy = m_Skeleton[i].y () - m_Skeleton[i].y0;
        const float dz = m_Skeleton[i].z () - m_Skeleton[i].z0;
        r += 0.01f * (dx * dx +
                      dy * dy +
                      dz * dz);
    }

    return r;
}

void BrushTuft::df (const float * x, float * result) const {
    float r = 0.0f;

    //return r;
}

void BrushTuft::relax (void) {
    float T = 0.2f;

    // Relax the nodes that aren't fixed.
    for (int iter = 256; iter; iter--) {
        bool converged = TRUE;

        // Setup the edges lengths.
        for (int i = 0; i < m_SkelLength; i++) {
            sTuftNode * node = m_Skeleton + i;
            for (int j = 0; j < node->nlinks; j++) {
                node->length1[j] = node->distanceTo (node->linkage[j]);
            }
        }

        // Go through nodes.
        for (int i = 0; i < m_SkelLength; i++) {
            sTuftNode * node = m_Skeleton + i;
            if (node->fixed)
                continue;

            float dx = 0.0f;
            float dy = 0.0f;
            float dz = 0.0f;

            // Estimate the tension..
            for (int j = 0; j < node->nlinks; j++) {
                const float dl = node->length1[j] - node->length0[j];
                if (fabsf (dl) > 0.001f) {
                    const sTuftNode * link = node->linkage[j];

                    const float x = (link->x () - node->x ()) / node->length1[j];
                    const float y = (link->y () - node->y ()) / node->length1[j];
                    const float z = (link->z () - node->z ()) / node->length1[j];

                    dx += T * dl * x;
                    dy += T * dl * y;
                    dz += T * dl * z;

                    converged = FALSE;
                }
            }

            // Compensate the position.
            node->setX (node->x () + dx);
            node->setY (node->y () + dy);
            node->setZ (node->z () + dz);
        }

        //T *= 0.95f;
        if (converged)
            break;
    }
}
*/
